import yaml
import collections


class OrderedYamlDict(yaml.YAMLObject, collections.OrderedDict):
    """A class which produces an ordered dict."""
    yaml_tag = "tag:yaml.org,2002:map"

    @classmethod
    def to_yaml(cls, dumper, data):
        value = []
        node = yaml.nodes.MappingNode(cls.yaml_tag, value)
        for key, item in data.iteritems():
            node_key = dumper.represent_data(key)
            node_value = dumper.represent_data(item)
            value.append((node_key, node_value))

        return node

    @classmethod
    def construct_mapping(cls, loader, node, deep=False):
        """Based on yaml.loader.BaseConstructor.construct_mapping."""

        if not isinstance(node, yaml.MappingNode):
            raise yaml.loader.ConstructorError(
                None, None, "expected a mapping node, but found %s" % node.id,
                node.start_mark)

        mapping = OrderedYamlDict()
        for key_node, value_node in node.value:
            key = loader.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise yaml.loader.ConstructorError(
                    "while constructing a mapping", node.start_mark,
                    "found unacceptable key (%s)" % exc, key_node.start_mark)

            value = loader.construct_object(value_node, deep=deep)
            mapping[key] = value

        return mapping

    @classmethod
    def from_yaml(cls, loader, node):
        """Parse the yaml file into an OrderedDict so we can preserve order."""
        fields = cls.construct_mapping(loader, node, deep=True)
        result = cls()
        for k, v in fields.items():
            result[k] = v

        return result


def decode(data):
    return yaml.safe_load(data) or OrderedYamlDict()

def encode(raw_data):
    return yaml.safe_dump(raw_data, default_flow_style=False)



# Rekall Memory Forensics
# Copyright 2015 Google Inc. All Rights Reserved.
#
# This file is part of Rekall Memory Forensics.
#
# Rekall Memory Forensics is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License Version 2 as
# published by the Free Software Foundation.  You may not use, modify or
# distribute this program under any other version of the GNU General Public
# License.
#
# Rekall Memory Forensics is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Rekall Memory Forensics.  If not, see <http://www.gnu.org/licenses/>.
#

"""Emulation of the Mimikatz tool.

This code replicates the algorithm first implemented in the mimikatz tool, which
can be found here:

https://github.com/gentilkiwi/mimikatz
"""

# pylint: disable=protected-access

__author__ = ("Michael Cohen <scudette@google.com> and "
              "Francesco Picasso <francesco.picasso@gmail.com>")

import logging

from Crypto.Cipher import AES
from Crypto.Cipher import DES3
from Crypto.Cipher import ARC4

from rekall import addrspace
from rekall import obj

from rekall.plugins.overlays.windows import pe_vtypes
from rekall.plugins.windows import common
from rekall.plugins.windows import lsadecryptxp


mimikatz_common_overlays = {
    '_LSA_UNICODE_STRING': [None, {
        'Value': lambda x: x.Buffer.dereference_as(
            'UnicodeString', target_args=dict(length=x.Length)),
        'Raw': lambda x: x.Buffer.dereference_as(
            'String', target_args=dict(length=x.Length)).v(),
        'RawMax': lambda x: x.Buffer.dereference_as(
            'String', target_args=dict(length=x.MaximumLength)).v(),
    }],
    '_LSA_STRING': [None, {
        'Value': lambda x: x.Buffer.dereference_as(
            'String', target_args=dict(length=x.Length)),
        'Raw': lambda x: x.Buffer.dereference_as(
            'String', target_args=dict(length=x.Length)).v(),
        'RawMax': lambda x: x.Buffer.dereference_as(
            'String', target_args=dict(length=x.MaximumLength)).v(),
    }],
    '_LUID': [None, {
        'Text': lambda x: '{:08x}:{:08x}'.format(x.HighPart, x.LowPart)
        }],
    '_SID': [None, {
        'IdentifierAuthority': [None, ['Enumeration', dict(
            choices={
                '\x00\x00\x00\x00\x00\x00': 'Null Authority',
                '\x00\x00\x00\x00\x00\x01': 'World Authority',
                '\x00\x00\x00\x00\x00\x02': 'Local Authority',
                '\x00\x00\x00\x00\x00\x03': 'Creator Authority',
                '\x00\x00\x00\x00\x00\x04': 'NonUnique Authority',
                '\x00\x00\x00\x00\x00\x05': 'NT Authority',
                },
            target='String',
            target_args=dict(length=6, term=None)
            )]],
        'NumericIdentifier': [0x4, ['unsigned be int']],
        'SubAuthority': [None, ['Array', dict(
            target='unsigned long',
            count=lambda x: x.SubAuthorityCount)]],
        }],
}


class _SID(obj.Struct):
    """A Pretty printing implementation of sids.

    Reference:
    http://www.sekchek.com/downloads/white-papers/windows-about-sids.pdf
    """
    def __unicode__(self):
        """Format the Sid using SDDL Notation."""
        components = [self.Revision, self.NumericIdentifier]
        components.extend(self.SubAuthority)

        return u"S-" + u"-".join([str(x) for x in components])


class Lsasrv(pe_vtypes.BasicPEProfile):
    """A profile for lsasrv.dll"""

    mimikatz_vtypes = [
        '_LIST_ENTRY', '_LSA_UNICODE_STRING', '_LUID',
        '_LSA_STRING', '_MSV1_0_PRIMARY_CREDENTIAL',
        '_KIWI_BCRYPT_HANDLE_KEY', '_KIWI_BCRYPT_KEY', '_KIWI_HARD_KEY',
        '_KIWI_MSV1_0_CREDENTIALS', '_KIWI_MSV1_0_PRIMARY_CREDENTIALS',
        '_KIWI_GENERIC_PRIMARY_CREDENTIAL',
        '_RPCE_CREDENTIAL_KEYCREDENTIAL', '_RPCE_COMMON_TYPE_HEADER',
        '_RPCE_PRIVATE_HEADER', '_MARSHALL_KEY',
        '_KIWI_MASTERKEY_CACHE_ENTRY', '_FILETIME']

    windows_vtypes = ['_SID', '_SID_IDENTIFIER_AUTHORITY', '_GUID']

    # TODO: should be special cases (1or2) addressed?
    mimikatz_msv_versioned = {
        5.1 : '_KIWI_MSV1_0_LIST_51',
        5.2 : '_KIWI_MSV1_0_LIST_52',
        6.0 : '_KIWI_MSV1_0_LIST_60',
        6.1 : '_KIWI_MSV1_0_LIST_61_ANTI_MIMIKATZ',
        6.2 : '_KIWI_MSV1_0_LIST_62',
        6.3 : '_KIWI_MSV1_0_LIST_63',
    }

    @classmethod
    def Initialize(cls, profile):
        super(cls, Lsasrv).Initialize(profile)

        arch = profile.session.profile.metadata('arch')

        mimikatz_profile = profile.session.LoadProfile('mimikatz/%s' % arch)
        if not mimikatz_profile:
            raise IOError('Unable to load mimikatz profile from repository!')

        kwargs = {}
        for name in cls.mimikatz_vtypes:
            kwargs[name] = mimikatz_profile.vtypes[name]

        for name in cls.windows_vtypes:
            kwargs[name] = profile.session.profile.vtypes[name]

        profile.add_types(kwargs)

        profile.add_types({
            'SIZED_DATA': [lambda x: x.size + 4, {
                'size': [0, ['unsigned long', {}]],
                'data': [4, ['String', dict(length=lambda x: x.size)]],
            }]
        })

        version = profile.session.profile.metadata('version')
        if version not in cls.mimikatz_msv_versioned:
            raise IOError('OS version not supported.')

        profile.add_types({
            'MSV1_0_LIST' : mimikatz_profile.vtypes[
                cls.mimikatz_msv_versioned[version]]
        })

        profile.add_classes(_SID=_SID)

        profile.add_overlay(mimikatz_common_overlays)

        profile.add_overlay({
            '_KIWI_HARD_KEY': [None, {
                'data': lambda x: x.m('data').cast(
                    'String', term=None, length=x.cbSecret)
            }],
            'MSV1_0_LIST': [None, {
                'List': [0, ['_LIST_ENTRY']],
                'pSid': [None, ['Pointer', dict(target='_SID')]],
                'LogonType': [None, ['Enumeration', dict(
                    target='unsigned int',
                    choices={
                        2: 'Interactive',
                        3: 'Network',
                        4: 'Batch',
                        5: 'Service',
                        6: 'Proxy',
                        7: 'Unlock',
                        8: 'NetworkCleartext',
                        9: 'NewCredentials',
                        10: 'RemoteInteractive',
                        11: 'CachedInteractive',
                        12: 'CachedRemoteInteractive',
                        13: 'CachedUnlock',
                    },
                )]],
            }],
            '_MSV1_0_PRIMARY_CREDENTIAL': [None, {
                'NtOwfPassword': [None, ['String', dict(length=16)]],
                'LmOwfPassword': [None, ['String', dict(length=16)]],
                'ShaOwPassword': [None, ['String', dict(length=20)]],
            }],
            '_MARSHALL_KEY': [None, {
                'unkId': [None, ['Enumeration', dict(
                    target='unsigned long',
                    choices={
                        0x00010002: 'NTLM',
                        0x00010003: 'NTLM',
                        0x00020002: 'SHA1',
                        0x00030002: 'RootKey',
                        0x00030003: 'RootKey',
                        0x00040002: 'DPAPI',
                        0x00040003: 'DPAPI',
                    },
                )]],
            }],
            '_RPCE_CREDENTIAL_KEYCREDENTIAL': [
                # TODO: the object size is wrong, data array takes the first
                # instance length, not all (different) lengths.
                lambda x: x.key.obj_offset + x.data.obj_size + x.key.obj_size,
                {
                    'key': [None, ['Array', {
                        'count': lambda x: x.unk0,
                        'target': '_MARSHALL_KEY'
                    }]],
                    'key_data': [lambda x: x.unk1.obj_end + x.key.obj_size, [
                        'Array', {
                            'count': lambda x: x.unk0,
                            'target': 'SIZED_DATA'
                        }]],
                }
            ],
            '_KIWI_MASTERKEY_CACHE_ENTRY': [None, {
                'List': [0, ['_LIST_ENTRY']],
                'key': [None, ['String', dict(length=lambda x: x.keySize)]],
            }],
        })

    def init_crypto(self):
        if self.session.profile.metadata('version') < 6.0:
            self.init_crypto_nt5()
        else:
            self.init_crypto_nt6()

    def decrypt(self, encrypted):
        if self.session.profile.metadata('version') < 6.0:
            return self.decrypt_nt5(encrypted)
        else:
            return self.decrypt_nt6(encrypted)

    def init_crypto_nt6(self):
        # TODO: add some checks to alert user if decryption is not possible.
        self.iv = self.get_constant_object(
            'InitializationVector', 'String', length=16, term=None).v()

        aes_handle = self.get_constant_object(
            'hAesKey', target='Pointer',
            target_args=dict(target='_KIWI_BCRYPT_HANDLE_KEY'))

        self.aes_key = aes_handle.key.hardkey.data.v()

        des_handle = self.get_constant_object(
            'h3DesKey', target='Pointer',
            target_args=dict(target='_KIWI_BCRYPT_HANDLE_KEY'))

        self.des_key = des_handle.key.hardkey.data.v()

    def decrypt_nt6(self, encrypted):
        cipher = None
        if self.iv:
            if len(encrypted) % 8:
                if self.aes_key:
                    cipher = AES.new(self.aes_key, AES.MODE_CFB, self.iv)
            else:
                if self.des_key:
                    cipher = DES3.new(self.des_key, DES3.MODE_CBC, self.iv[:8])
        if cipher and encrypted:
            return cipher.decrypt(encrypted)
        return obj.NoneObject()

    def init_crypto_nt5(self):
        # TODO: add some checks to alert user if decryption is not possible.
        rc4_key_len = self.get_constant_object(
            'g_cbRandomKey', 'unsigned long').v()

        rc4_key_ptr = self.get_constant_object(
            'g_pRandomKey', target='Pointer')

        self.rc4_key = rc4_key_ptr.dereference_as(
            'String', target_args=dict(length=rc4_key_len, term=None)).v()

        desx_key_ptr = self.get_constant_object(
            'g_pDESXKey', target='Pointer')

        self.desx_key = desx_key_ptr.dereference_as(
            'String', target_args=dict(length=144, term=None)).v()

        self.feedback = self.get_constant_object(
            'g_Feedback', target='String',
            target_args=dict(length=8)).v()

    def decrypt_nt5(self, encrypted):
        cipher = None
        if len(encrypted) % 8:
            if self.rc4_key:
                cipher = ARC4.new(self.rc4_key)
        else:
            if self.desx_key and self.feedback:
                cipher = lsadecryptxp.XP_LsaDecryptMemory(
                    self.desx_key, self.feedback)
        if cipher and encrypted:
            return cipher.decrypt(encrypted)
        return obj.NoneObject()

    def get_lsass_logons(self):
        logons = {}
        lsass_logons = self.get_constant_object(
            'LogonSessionList', target='_LIST_ENTRY')
        for entry in lsass_logons.list_of_type('MSV1_0_LIST', 'List'):
            logons[entry.LocallyUniqueIdentifier.Text] = entry
        return logons

    def _msv_primary_credentials(self, data):
        vm = addrspace.BufferAddressSpace(data=data, session=self.session)
        cred_obj = self.Object('_MSV1_0_PRIMARY_CREDENTIAL',
                               profile=self, vm=vm)

        # TODO: check NULL Pointer dereference with this VM.
        domain = ''
        if cred_obj.LogonDomainName.Buffer.is_valid():
            domain = cred_obj.LogonDomainName.Value

        user_name = ''
        if cred_obj.UserName.Buffer.is_valid():
            user_name = cred_obj.UserName.Value

        if cred_obj.isLmOwfPassword.v() == 1:
            yield (domain, user_name, 'LM',
                   cred_obj.LmOwfPassword.v().encode('hex'))

        if cred_obj.isNtOwfPassword.v() == 1:
            yield (domain, user_name, 'NTLM',
                   cred_obj.NtOwfPassword.v().encode('hex'))

        if cred_obj.isShaOwPassword.v() == 1:
            yield (domain, user_name, 'SHA1',
                   cred_obj.ShaOwPassword.v().encode('hex'))

    def _msv_rpce_credentials(self, data):
        vm = addrspace.BufferAddressSpace(data=data, session=self.session)
        cred_obj = self.Object('_RPCE_CREDENTIAL_KEYCREDENTIAL',
                               profile=self, vm=vm)

        # This seems to be corrupt sometimes.
        if cred_obj.unk0 > 10:
            return

        for i in range(0, cred_obj.unk0):
            yield (cred_obj.key[i].unkId,
                   cred_obj.key_data[i].data.v().encode('hex'))

    def logons(self, lsass_logons):
        for luid, lsass_logon in lsass_logons.iteritems():
            for cred in lsass_logon.Credentials.walk_list('next'):
                for primary_cred in cred.PrimaryCredentials.walk_list('next'):

                    dec_cred = self.decrypt(primary_cred.Credentials.Raw)
                    if not dec_cred:
                        continue

                    cur_cred_type = primary_cred.Primary.Value

                    if cur_cred_type == u'Primary':
                        for (domain, user_name, secret_type,
                             secret) in self._msv_primary_credentials(dec_cred):
                            yield (luid, cur_cred_type, domain, user_name,
                                   secret_type, secret)

                    elif cur_cred_type == u'CredentialKeys':
                        for (secret_type,
                             secret) in self._msv_rpce_credentials(dec_cred):
                            yield (luid, cur_cred_type, '', '', secret_type,
                                   secret)
                    else:
                        pass

    def master_keys(self):
        keys = self.get_constant_object(
            'g_MasterKeyCacheList', target='_LIST_ENTRY')
        for entry in keys.list_of_type('_KIWI_MASTERKEY_CACHE_ENTRY', 'List'):
            yield (entry.LogonId.Text, '', # TODO: add entry.KeyUid,
                   '', '', 'masterkey',
                   self.decrypt(entry.key.v()).encode('hex'))


class Wdigest(pe_vtypes.BasicPEProfile):
    """A profile for wdigest.dll"""

    mimikatz_vtypes = [
        '_LIST_ENTRY', '_LSA_UNICODE_STRING', '_LUID',
        '_KIWI_WDIGEST_LIST_ENTRY', '_KIWI_GENERIC_PRIMARY_CREDENTIAL',
        '_KIWI_HARD_KEY']

    @classmethod
    def Initialize(cls, profile):
        super(cls, Wdigest).Initialize(profile)

        arch = profile.session.profile.metadata('arch')
        mimikatz_profile = profile.session.LoadProfile('mimikatz/%s' % arch)

        kwargs = {}
        for name in cls.mimikatz_vtypes:
            kwargs[name] = mimikatz_profile.vtypes[name]

        profile.add_types(kwargs)

        profile.add_overlay(mimikatz_common_overlays)

        kiwi_cred_offset = 8
        if profile.session.profile.metadata('version') < 6.0:
            kiwi_cred_offset = 12

        profile.add_overlay({
            '_KIWI_WDIGEST_LIST_ENTRY': [None, {
                'List': [0, ['_LIST_ENTRY']],
                'Cred': [lambda x: (x.LocallyUniqueIdentifier.obj_end +
                                    kiwi_cred_offset),
                         ['_KIWI_GENERIC_PRIMARY_CREDENTIAL']]
            }],
            '_KIWI_HARD_KEY': [None, {
                'data': lambda x: x.m('data').cast(
                    'String', term=None, length=x.cbSecret)
            }],
        })

    def logons(self):
        # TODO: if the symbols is wrong? Add a check for the LIST validity.
        logons = self.get_constant_object(
            'l_LogSessList', target='_LIST_ENTRY')
        for entry in logons.list_of_type('_KIWI_WDIGEST_LIST_ENTRY', 'List'):
            yield entry


class Livessp(pe_vtypes.BasicPEProfile):
    """A profile for livessp.dll"""

    mimikatz_vtypes = [
        '_LIST_ENTRY', '_LSA_UNICODE_STRING', '_LUID',
        '_KIWI_LIVESSP_LIST_ENTRY', '_KIWI_LIVESSP_PRIMARY_CREDENTIAL',
        '_KIWI_GENERIC_PRIMARY_CREDENTIAL']

    @classmethod
    def Initialize(cls, profile):
        super(cls, Livessp).Initialize(profile)

        arch = profile.session.profile.metadata('arch')
        mimikatz_profile = profile.session.LoadProfile('mimikatz/%s' % arch)

        kwargs = {}
        for name in cls.mimikatz_vtypes:
            kwargs[name] = mimikatz_profile.vtypes[name]

        profile.add_types(kwargs)

        profile.add_overlay(mimikatz_common_overlays)

        profile.add_overlay({
            '_KIWI_LIVESSP_LIST_ENTRY': [None, {
                'List': [0, ['_LIST_ENTRY']],
            }]
        })

    def logons(self):
        logons = self.get_constant_object(
            'LiveGlobalLogonSessionList', target='_LIST_ENTRY')
        for entry in logons.list_of_type('_KIWI_LIVESSP_LIST_ENTRY', 'List'):
            yield (entry.LocallyUniqueIdentifier.Text,
                   '',
                   entry.suppCreds.dereference().credentials.Domaine.Value,
                   entry.suppCreds.dereference().credentials.UserName.Value,
                   'password',
                   entry.suppCreds.dereference().credentials.Password)


class Mimikatz(common.WindowsCommandPlugin):
    """Extract and decrypt passwords from the LSA Security Service."""

    name = 'mimikatz'

    def __init__(self, **kwargs):
        super(Mimikatz, self).__init__(**kwargs)

        # Track the following modules. If we do not have them in the profile
        # repository then try to get them directly from Microsoft.
        tracked = self.session.GetParameter(
            'autodetect_build_local_tracked') or []

        needed = set(['lsasrv', 'wdigest', 'livessp'])
        if not needed.issubset(tracked):
            needed.update(tracked)
            with self.session as session:
                session.SetParameter('autodetect_build_local_tracked', needed)

    def render(self, renderer):
        renderer.table_header([
            dict(name='LUID', width=20),
            dict(name='Type', width=16),
            dict(name='Sess', width=2),
            dict(name='SID', width=20),
            dict(name='Module', width=7),
            dict(name='Info', width=7),
            dict(name='Domain', width=16),
            dict(name='User', width=16),
            dict(name='SType', width=9),
            dict(name='Secret', width=32)])

        cc = self.session.plugins.cc()
        # Switch to the lsass process.
        for task in self.session.plugins.pslist(
                proc_regex='lsass.exe').filter_processes():
            cc.SwitchProcessContext(task)

            lsasrv = self.session.address_resolver.LoadProfileForName(
                'lsasrv')

            lsasrv.init_crypto()
            lsass_logons = lsasrv.get_lsass_logons()

            for (luid, info, domain, user_name, secret_type,
                 secret) in  lsasrv.logons(lsass_logons):

                lsass_entry = lsass_logons.get(luid, obj.NoneObject())
                # TODO: add timestamp field?
                row = (luid,
                       lsass_entry.LogonType,
                       lsass_entry.Session,
                       lsass_entry.pSid.deref(),
                       'msv',
                       info,
                       domain,
                       user_name,
                       secret_type,
                       secret)

                renderer.table_row(*row)

            wdigest = self.session.address_resolver.LoadProfileForName(
                'wdigest')

            if not wdigest.get_constant('l_LogSessList'):
                logging.warning('wdigest not initializated, skipping it.')
            else:
                for entry in wdigest.logons():
                    luid = entry.LocallyUniqueIdentifier.Text
                    lsass_entry = lsass_logons.get(luid, obj.NoneObject())

                    row = (luid,
                           lsass_entry.LogonType,
                           lsass_entry.Session,
                           lsass_entry.pSid.deref(),
                           'wdigest',
                           '',
                           entry.Cred.Domaine.Value,
                           entry.Cred.UserName.Value,
                           'password',
                           lsasrv.decrypt(entry.Cred.Password.RawMax))

                    renderer.table_row(*row)

            livessp = self.session.address_resolver.LoadProfileForName(
                'livessp')

            if not livessp.get_constant('LiveGlobalLogonSessionList'):
                logging.warning('livessp not initializated, skipping it.')
            else:
                for (luid, info, domain, user_name, secret_type,
                     enc_secret) in livessp.logons():
                    lsass_entry = lsass_logons.get(luid, obj.NoneObject())

                    row = (luid,
                           lsass_entry.LogonType,
                           lsass_entry.Session,
                           lsass_entry.pSid.deref(),
                           'livessp',
                           info,
                           domain,
                           user_name,
                           secret_type,
                           lsasrv.decrypt(enc_secret))

                    renderer.table_row(*row)

            for (luid, info, domain, user_name, secret_type,
                 secret) in lsasrv.master_keys():
                lsass_entry = lsass_logons.get(luid, obj.NoneObject())

                row = (luid,
                       lsass_entry.LogonType,
                       lsass_entry.Session,
                       lsass_entry.pSid.deref(),
                       'lsasrv',
                       info,
                       domain,
                       user_name,
                       secret_type,
                       secret)

                renderer.table_row(*row)



# Rekall Memory Forensics
#
# Copyright 2012 Philippe Teuwen, Thorsten Sick, Michael Cohen
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""An Address Space for processing ELF64 coredumps."""
# References:
# VirtualBox core format:
# http://www.virtualbox.org/manual/ch12.html#guestcoreformat
# ELF64 format: http://downloads.openwatcom.org/ftp/devel/docs/elf-64-gen.pdf

# Note that as of version 1.6.0 WinPmem also uses ELF64 as the default imaging
# format. Except that WinPmem stores image metadata in a YAML file stored in the
# image. This address space supports both formats.
import os
import yaml

from rekall import addrspace
from rekall.plugins.overlays.linux import elf

PT_PMEM_METADATA = 0x6d656d70  # Spells 'pmem'



class Elf64CoreDump(addrspace.RunBasedAddressSpace):
    """This AS supports ELF64 coredump format, as used by VirtualBox."""
    order = 30

    __name = "elf64"
    __image = True

    def __init__(self, **kwargs):
        super(Elf64CoreDump, self).__init__(**kwargs)

        # Check the file for sanity.
        self.check_file()

        self.offset = 0
        self.fname = ''
        self._metadata = {}

        # Now parse the ELF file.
        elf_profile = elf.ELFProfile(session=self.session)
        self.elf64_hdr = elf_profile.elf64_hdr(vm=self.base, offset=0)

        self.as_assert(self.elf64_hdr.e_type == "ET_CORE",
                       "Elf file is not a core file.")
        self.name = "%s|%s" % (self.__class__.__name__, self.base.name)

        # Iterate over all the program headers and map the runs.
        for segment in self.elf64_hdr.e_phoff:
            if segment.p_type == "PT_LOAD":
                # Some load segments are empty.
                if (segment.p_filesz == 0 or
                        segment.p_filesz != segment.p_memsz):
                    continue

                # Add the run to the memory map.
                virtual_address = int(segment.p_paddr) or int(segment.p_vaddr)
                self.runs.insert((virtual_address,  # Virtual Addr
                                  int(segment.p_offset), # File Addr
                                  int(segment.p_memsz))) # Length

            elif segment.p_type == PT_PMEM_METADATA:
                self.LoadMetadata(segment.p_offset)

        # Search for the pmem footer signature.
        footer = self.base.read(self.base.end() - 10000, 10000)
        if "...\n" in footer[-6:]:
            header_offset = footer.rfind("# PMEM")
            if header_offset > 0:
                self.LoadMetadata(self.base.end() - 10000 + header_offset)

    def check_file(self):
        """Checks the base file handle for sanity."""

        self.as_assert(self.base,
                       "Must stack on another address space")

        ## Must start with the magic for elf
        self.as_assert((self.base.read(0, 4) == "\177ELF"),
                       "Header signature invalid")

    def LoadMetadata(self, offset):
        """Load the WinPmem metadata from the elf file."""
        try:
            data = self.base.read(offset, 1024*1024)
            yaml_file = data.split('...\n')[0]

            metadata = yaml.safe_load(yaml_file)
        except (yaml.YAMLError, TypeError) as e:
            self.session.logging.error(
                "Invalid file metadata, skipping: %s" % e)
            return

        for session_param, metadata_key in (("dtb", "CR3"),
                                            ("kernel_base", "KernBase")):
            if metadata_key in metadata:
                self.session.SetParameter(
                    session_param, metadata[metadata_key])

        previous_section = metadata.pop("PreviousHeader", None)
        if previous_section is not None:
            self.LoadMetadata(previous_section)

        pagefile_offset = metadata.get("PagefileOffset", None)
        pagefile_size = metadata.get("PagefileSize", None)

        if pagefile_offset is not None and pagefile_size is not None:
            self.LoadPageFile(pagefile_offset, pagefile_size)

        self._metadata.update(metadata)

    pagefile_offset = 0
    pagefile_end = 0

    def LoadPageFile(self, pagefile_offset, pagefile_size):
        """We map the page file into the physical address space.

        This allows us to treat all physical addresses equally - regardless if
        they come from the memory or the page file.
        """
        # Map the pagefile after the end of the physical address space.
        vaddr = self.end() + 0x10000

        self.session.logging.info(
            "Loading pagefile into physical offset %#08x", vaddr)
        self.runs.insert((vaddr, pagefile_offset, pagefile_size))

        # Remember the region for the pagefile.
        self.pagefile_offset = vaddr
        self.pagefile_end = vaddr + pagefile_size

    def describe(self, addr):
        if self.pagefile_offset <= addr <= self.pagefile_end:
            return "%#x@Pagefile" % (
                addr - self.pagefile_offset)

        return "%#x" % addr


class KCoreAddressSpace(Elf64CoreDump):
    """A Linux kernel's /proc/kcore file also maps the entire physical ram.

    http://lxr.free-electrons.com/source/Documentation/x86/x86_64/mm.txt

    ffff880000000000 - ffffc7ffffffffff (=64 TB) direct mapping of all
    physical memory.
    """
    # We must run before the regular Elf64CoreDump address space in the voting
    # order.
    order = Elf64CoreDump.order - 1

    __name = "elf64"
    __image = True

    volatile = True

    def __init__(self, **kwargs):
        super(KCoreAddressSpace, self).__init__(**kwargs)

        # Collect all ranges between ffff880000000000 - ffffc7ffffffffff
        runs = []

        for vaddr, paddr, length in self.runs:
            if 0xffff880000000000 < vaddr < 0xffffc7ffffffffff:
                runs.append((vaddr - 0xffff880000000000, paddr, length))

        self.as_assert(runs, "No kcore compatible virtual ranges.")
        self.runs.clear()

        # At this point, we think this is a valid, usable kcore file.
        # RHEL, however, disabled read access to /proc/kcore past the ELF
        # headers and the file size reflects this. /proc/kcore usually has a
        # size of at least 64TB (46bits of physical address space in x64).
        # We use the file size to detect cases where kcore will be unusable.

        if getattr(self.base, "fhandle", None):
            try:
                statinfo = os.fstat(self.base.fhandle.fileno())
                if statinfo.st_size < 2**46:
                    # We raise a TypeError and not an ASAssertionError because
                    # we need it to be catchable when rekall is used as a
                    # library (i.e GRR). ASAssertionErrors are swallowed by
                    # the address space selection algorithm.
                    raise RuntimeError(
                        "This kcore file is too small (%d bytes) and likely "
                        "invalid for memory analysis. You may want to use pmem "
                        "instead." % statinfo.st_size)
            except(IOError, AttributeError):
                pass

        for x in runs:
            self.runs.insert(x)


def WriteElfFile(address_space, outfd, session=None):
    """Convert the address_space to an ELF Core dump file.

    The Core dump will be written to outfd which is expected to have a .write()
    method.
    """
    runs = list(address_space.get_available_addresses())

    elf_profile = elf.ELFProfile(session=session)
    elf64_pheader = elf_profile.elf64_phdr()
    elf64_pheader.p_type = "PT_LOAD"
    elf64_pheader.p_align = 0x1000
    elf64_pheader.p_flags = "PF_R"

    elf64_header = elf_profile.elf64_hdr()
    elf64_header.e_ident = elf64_header.e_ident.signature
    elf64_header.e_type = 'ET_CORE'
    elf64_header.e_phoff = elf64_header.obj_end
    elf64_header.e_ehsize = elf64_header.obj_size
    elf64_header.e_phentsize = elf64_pheader.obj_size
    elf64_header.e_phnum = len(runs)
    elf64_header.e_shnum = 0  # We don't have any sections.

    # Where we start writing data.
    file_offset = (elf64_header.obj_size +
                   # One Phdr for each run.
                   len(runs) * elf64_pheader.obj_size)

    outfd.write(elf64_header.GetData())
    for offset, _, length in runs:
        elf64_pheader.p_paddr = offset
        elf64_pheader.p_memsz = length
        elf64_pheader.p_offset = file_offset
        elf64_pheader.p_filesz = length

        outfd.write(elf64_pheader.GetData())

        file_offset += length

    # Now just copy all the runs
    total_data = 0
    for offset, _, length in runs:
        while length > 0:
            data = address_space.read(offset, min(10000000, length))
            session.report_progress("Writing %sMb", total_data/1024/1024)
            outfd.write(data)
            length -= len(data)
            offset += len(data)
            total_data += len(data)



#!/usr/bin/python

# Rekall Memory Forensics
# Copyright (C) 2012 Michael Cohen <scudette@gmail.com>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This profile is for the MACH-O file format.

For now this is not a complete implementation - just enough to parse the MACHO
core dump style images.

References:
https://developer.apple.com/library/mac/#documentation/developertools/conceptual/MachORuntime/Reference/reference.html

http://www.opensource.apple.com/source/xnu/xnu-1456.1.26/EXTERNAL_HEADERS/mach-o/loader.h

http://svn.red-bean.com/bob/macholib/trunk/macholib/mach_o.py
"""

__author__ = "Michael Cohen <scudette@gmail.com>"

from rekall import obj
from rekall.plugins.overlays import basic


macho_vtypes = {
 'mach_header_64': [ 0x20, {
    'magic': [0x0, ['unsigned int']],
    'cputype': [0x4, ['Enumeration', dict(
                        choices={
                            1:      'VAX',
                            6:      'MC680x0',
                            7:      'i386',
                            8:      'MIPS',
                            10:     'MC98000',
                            11:     'HPPA',
                            12:     'ARM',
                            13:     'MC88000',
                            14:     'SPARC',
                            15:     'i860',
                            16:     'Alpha',
                            18:     'PowerPC',
                            (0x01000000 + 7): 'X86_64',
                            (0x01000000 + 18): 'PowerPC_64',
                            },
                        target="unsigned int",
                        )]],
    'cpusubtype': [0x8, ['int']],
    'filetype': [0xc, ['Enumeration', dict(
                        choices={
                            0x1: 'MH_OBJECT',  # relocatable object file
                            0x2: "MH_EXECUTE", # demand paged executable file
                            0x3: "MH_FVMLIB",  # fixed VM shared library file
                            0x4: 'MH_CORE',    # core file
                            0x5: 'MH_PRELOAD', # preloaded executable file
                            0x6: 'MH_DYLIB',   # dynamicly bound shared library file
                            0x7: 'MH_DYLINKER',# dynamic link editor
                            0x8: 'MH_BUNDLE',  # dynamicly bound bundle file
                            },
                        target="unsigned int",
                        )]],
    'ncmds': [0x10, ['unsigned int']],
    'sizeofcmds': [0x14, ['unsigned int']],
    'flags': [0x18, ['Flags', dict(maskmap={
                        'MH_NOUNDEFS':     0x1,
                        'MH_INCRLINK':     0x2,
                        'MH_DYLDLINK':     0x4,
                        'MH_BINDATLOAD':   0x8,
                        'MH_PREBOUND':     0x10,
                        })]],
    'reserved': [0x1c, ['unsigned int']],
    'segments': [0x20, ['Array', dict(
                        target="segment_command_64",
                        count=lambda x: x.ncmds)]],
}],

'segment_command_64': [ 0x48, {
    'cmd': [0x0, ['Enumeration', dict(
                        choices={
                            0x1: 'LC_SEGMENT',
                            0x2: 'LC_SYMTAB',
                            0x3: 'LC_SYMSEG',
                            0x4: 'LC_THREAD',
                            0x5: 'LC_UNIXTHREAD',
                            0x6: 'LC_LOADFVMLIB',
                            0x7: 'LC_IDFVMLIB',
                            0x8: 'LC_IDENT',
                            0x9: 'LC_FVMFILE',
                            0xa: 'LC_PREPAGE',
                            0xb: 'LC_DYSYMTAB',
                            0xc: 'LC_LOAD_DYLIB',
                            0xd: 'LC_ID_DYLIB',
                            0xe: 'LC_LOAD_DYLINKER',
                            0xf: 'LC_ID_DYLINKER',
                            0x10: 'LC_PREBOUND_DYLIB',
                            0x11: 'LC_ROUTINES',
                            0x12: 'LC_SUB_FRAMEWORK',
                            0x13: 'LC_SUB_UMBRELLA',
                            0x14: 'LC_SUB_CLIENT',
                            0x15: 'LC_SUB_LIBRARY',
                            0x16: 'LC_TWOLEVEL_HINTS',
                            0x17: 'LC_PREBIND_CKSUM',
                            0x80000000 + 0x18: 'LC_LOAD_WEAK_DYLIB',
                            0x19: 'LC_SEGMENT_64',
                            0x1a: 'LC_ROUTINES_64',
                            0x1b: 'LC_UUID',
                            0x80000000 + 0x1c: 'LC_RPATH',
                            0x1d: 'LC_CODE_SIGNATURE',
                            0x1e: 'LC_SEGMENT_SPLIT_INFO',
                            0x80000000 + 0x1f: 'LC_REEXPORT_DYLIB',
                            0x20: 'LC_LAZY_LOAD_DYLIB',
                            0x21: 'LC_ENCRYPTION_INFO',
                            0x22: 'LC_DYLD_INFO',
                            0x80000000 + 0x22: 'LC_DYLD_INFO_ONLY',
                            },
                        target="unsigned int")]],
    'cmdsize': [0x4, ['unsigned int']],
    'segname': [0x8, ['String', dict(length=16)]],
    'vmaddr': [0x18, ['unsigned long long']],
    'vmsize': [0x20, ['unsigned long long']],
    'fileoff': [0x28, ['unsigned long long']],
    'filesize': [0x30, ['unsigned long long']],
    'maxprot': [0x38, ['int']],
    'initprot': [0x3c, ['int']],
    'nsects': [0x40, ['unsigned int']],
    'flags': [0x44, ['unsigned int']],
}],
}


class MACHOFileImplementation(obj.ProfileModification):
    """An implementation of a parser for MAC-O files."""

    @classmethod
    def Modify(cls, profile):
        profile.add_types(macho_vtypes)


class MACHO64Profile(basic.ProfileLP64, basic.BasicClasses):
    """A profile for MAC-O files."""

    def __init__(self, **kwargs):
        super(MACHO64Profile, self).__init__(**kwargs)

        MACHOFileImplementation.Modify(self)




# Rekall Memory Forensics
# Copyright (C) 2008 Volatile Systems
# Copyright (C) 2007, 2008 Brendan Dolan-Gavitt bdolangavitt@wesleyan.edu
# http://moyix.blogspot.com
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Additional Authors:
# Mike Auty <mike.auty@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# Based heavily upon the getsids plugin by Moyix
# http://kurtz.cs.wesleyan.edu/%7Ebdolangavitt/memory/getsids.py

"""
@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""

from rekall.plugins.windows import common
import re

class GetSIDs(common.WinProcessFilter):
    """Print the SIDs owning each process token."""

    __name = "tokens"

    well_known_sid_re = [
      (re.compile(r'S-1-5-[0-9-]+-500'), 'Administrator'),
      (re.compile(r'S-1-5-[0-9-]+-501'), 'Guest'),
      (re.compile(r'S-1-5-[0-9-]+-502'), 'KRBTGT'),
      (re.compile(r'S-1-5-[0-9-]+-512'), 'Domain Admins'),
      (re.compile(r'S-1-5-[0-9-]+-513'), 'Domain Users'),
      (re.compile(r'S-1-5-[0-9-]+-514'), 'Domain Guests'),
      (re.compile(r'S-1-5-[0-9-]+-515'), 'Domain Computers'),
      (re.compile(r'S-1-5-[0-9-]+-516'), 'Domain Controllers'),
      (re.compile(r'S-1-5-[0-9-]+-517'), 'Cert Publishers'),
      (re.compile(r'S-1-5-[0-9-]+-520'), 'Group Policy Creator Owners'),
      (re.compile(r'S-1-5-[0-9-]+-533'), 'RAS and IAS Servers'),
      (re.compile(r'S-1-5-5-[0-9]+-[0-9]+'), 'Logon Session'),
      (re.compile(r'S-1-5-21-[0-9-]+-518'), 'Schema Admins'),
      (re.compile(r'S-1-5-21-[0-9-]+-519'), 'Enterprise Admins'),
      (re.compile(r'S-1-5-21-[0-9-]+-553'), 'RAS Servers'),
    ]

    well_known_sids = {
      'S-1-0': 'Null Authority',
      'S-1-0-0': 'Nobody',
      'S-1-1': 'World Authority',
      'S-1-1-0': 'Everyone',
      'S-1-2': 'Local Authority',
      'S-1-2-0': 'Local (Users with the ability to log in locally)',
      'S-1-2-1': ('Console Logon (Users who are logged onto the physical '
                  'console)'),
      'S-1-3': 'Creator Authority',
      'S-1-3-0': 'Creator Owner',
      'S-1-3-1': 'Creator Group',
      'S-1-3-2': 'Creator Owner Server',
      'S-1-3-3': 'Creator Group Server',
      'S-1-3-4': 'Owner Rights',
      'S-1-4': 'Non-unique Authority',
      'S-1-5': 'NT Authority',
      'S-1-5-1': 'Dialup',
      'S-1-5-2': 'Network',
      'S-1-5-3': 'Batch',
      'S-1-5-4': 'Interactive',
      'S-1-5-6': 'Service',
      'S-1-5-7': 'Anonymous',
      'S-1-5-8': 'Proxy',
      'S-1-5-9': 'Enterprise Domain Controllers',
      'S-1-5-10': 'Principal Self',
      'S-1-5-11': 'Authenticated Users',
      'S-1-5-12': 'Restricted Code',
      'S-1-5-13': 'Terminal Server Users',
      'S-1-5-14': 'Remote Interactive Logon',
      'S-1-5-15': 'This Organization',
      'S-1-5-17': 'This Organization (Used by the default IIS user)',
      'S-1-5-18': 'Local System',
      'S-1-5-19': 'NT Authority',
      'S-1-5-20': 'NT Authority',
      'S-1-5-32-544': 'Administrators',
      'S-1-5-32-545': 'Users',
      'S-1-5-32-546': 'Guests',
      'S-1-5-32-547': 'Power Users',
      'S-1-5-32-548': 'Account Operators',
      'S-1-5-32-549': 'Server Operators',
      'S-1-5-32-550': 'Print Operators',
      'S-1-5-32-551': 'Backup Operators',
      'S-1-5-32-552': 'Replicators',
      'S-1-5-32-554': 'BUILTIN\\Pre-Windows 2000 Compatible Access',
      'S-1-5-32-555': 'BUILTIN\\Remote Desktop Users',
      'S-1-5-32-556': 'BUILTIN\\Network Configuration Operators',
      'S-1-5-32-557': 'BUILTIN\\Incoming Forest Trust Builders',
      'S-1-5-32-558': 'BUILTIN\\Performance Monitor Users',
      'S-1-5-32-559': 'BUILTIN\\Performance Log Users',
      'S-1-5-32-560': 'BUILTIN\\Windows Authorization Access Group',
      'S-1-5-32-561': 'BUILTIN\\Terminal Server License Servers',
      'S-1-5-32-562': 'BUILTIN\\Distributed COM Users',
      'S-1-5-32-568': 'BUILTIN\\IIS IUSRS',
      'S-1-5-32-569': 'Cryptographic Operators',
      'S-1-5-32-573': 'BUILTIN\\Event Log Readers',
      'S-1-5-32-574': 'BUILTIN\\Certificate Service DCOM Access',
      'S-1-5-33': 'Write Restricted',
      'S-1-5-64-10': 'NTLM Authentication',
      'S-1-5-64-14': 'SChannel Authentication',
      'S-1-5-64-21': 'Digest Authentication',
      'S-1-5-80': 'NT Service',
      'S-1-5-86-1544737700-199408000-2549878335-3519669259-381336952': (
          'WMI (Local Service)'),
      'S-1-5-86-615999462-62705297-2911207457-59056572-3668589837': (
          'WMI (Network Service)'),
      'S-1-5-1000': 'Other Organization',
      'S-1-16-0': 'Untrusted Mandatory Level',
      'S-1-16-4096': 'Low Mandatory Level',
      'S-1-16-8192': 'Medium Mandatory Level',
      'S-1-16-8448': 'Medium Plus Mandatory Level',
      'S-1-16-12288': 'High Mandatory Level',
      'S-1-16-16384': 'System Mandatory Level',
      'S-1-16-20480': 'Protected Process Mandatory Level',
      'S-1-16-28672': 'Secure Process Mandatory Level',
    }

    def find_sid_re(self, sid_string):
        for reg, name in self.well_known_sid_re:
            if reg.search(sid_string):
                return name

    def render(self, renderer):
        """Renders the sids as text"""
        # First enumerate all the users on this system.
        users = {}
        for _, v, f in self.session.plugins.users().GenerateUsers():
            users[f.Rid] = v.UserName.Value

        renderer.table_header([
            ("Process", "process", "16"),
            ("Pid", "pid", "5"),
            ("Sid", "sid", "50"),
            ("Comment", "comment", "")
            ])

        for task in self.filter_processes():
            for sa in task.Token.UserAndGroups:
                # The RID is the last SubAuthority
                rid = sa.Sid.SubAuthority[sa.Sid.SubAuthorityCount-1]
                sid_name = ""
                username = users.get(rid)
                if username:
                    sid_name = "User: %s" % username

                sid_string = unicode(sa.Sid.deref())
                if sid_string in self.well_known_sids:
                    sid_name = self.well_known_sids[sid_string]
                else:
                    sid_name_re = self.find_sid_re(sid_string)
                    if sid_name_re:
                        sid_name = sid_name_re

                renderer.table_row(
                    task.ImageFileName, task.UniqueProcessId,
                    sid_string, sid_name)



# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Original Source:
# Copyright (C) 2004,2005,2006 4tphi Research
# Author: {npetroni,awalters}@4tphi.net (Nick Petroni and AAron Walters)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       AAron Walters
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com
@organization: Volatile Systems

   Alias for all address spaces

"""
import weakref

from rekall import registry
from rekall import utils


class TranslationLookasideBuffer(utils.FastStore):
    """An implementation of a TLB.

    This can be used by an address space to cache translations.
    """

    PAGE_SHIFT = 12
    PAGE_MASK = ~ 0xFFF

    def Get(self, vaddr):
        result = super(TranslationLookasideBuffer, self).Get(
            vaddr >> self.PAGE_SHIFT)

        if result is not None:
            return result + (vaddr & 0xFFF)

    def Put(self, vaddr, paddr):
        if paddr is not None:
            paddr = paddr & self.PAGE_MASK

        super(TranslationLookasideBuffer, self).Put(
            vaddr >> self.PAGE_SHIFT, paddr)


class BaseAddressSpace(object):
    """ This is the base class of all Address Spaces. """

    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    order = 10

    # This can be used to name the address space (e.g. process if etc).
    name = ""

    # Some useful metadata for address spaces.

    # This signifies that this address space normally operates on memory
    # images. This flag controls if this address space will participate in
    # address space autoselection for image detection. Note that it can not be
    # inherited but must be explicitly set.
    __image = False

    # This flag signifies whether this address space's contents are likely to
    # change between reads. If an address space is NOT volatile (this flag is
    # False) then reads from the same offset MUST always return the same bytes.
    volatile = False

    def __init__(self, base=None, session=None, profile=None, **_):
        """Base is the AS we will be stacking on top of, opts are options which
        we may use.

        Args:
          base: A base address space to stack on top of (i.e. delegate to it for
              satisfying read requests).

          session: An optional session object.

          profile: An optional profile to use for parsing the address space
              (e.g. needed for hibernation, crash etc.)
        """
        if session is None and base is not None:
            session = base.session

        self.base = base
        if base:
            self.volatile = self.base.volatile

        # This is the base address space which this address space reads from. In
        # this context, this address space is used to read physical addresses as
        # obtained from get_available_addresses(). NOTE: This is a weak
        # reference to avoid creating a cycle.
        self.phys_base = weakref.proxy(self)

        self.profile = profile
        self.session = session
        if session is None:
            raise RuntimeError("Session must be provided.")

        # This is a short lived cache. If we use a static image, this cache need
        # not expire, however, when analysing a live system we need to flush the
        # cache frequently.
        self.cache = utils.AgeBasedCache(max_age=20)

    def as_assert(self, assertion, error=None):
        """Duplicate for the assert command (so that optimizations don't disable
        them)

        It had to be called as_assert, since assert is a keyword
        """
        if not assertion:
            raise ASAssertionError(
                error or "Instantiation failed for unspecified reason")

    def describe(self, addr):
        """Return a string describing an address."""
        return "%#x" % addr

    def read(self, unused_addr, length):
        """Should be overridden by derived classes."""
        if length > self.session.GetParameter("buffer_size"):
            raise IOError("Too much data to read.")

        return "\x00" * length

    def get_available_addresses(self, start=0):
        """Generates address ranges (offset, phys_offset, size) for this AS.

        NOTE!! The phys_offset here refers to an address read from the phys_base
        member of this address space. I.e. it should be possible for callers to
        call get_available_addresses() and then directly get data via
        address_space.phys_base.read(phys_offset, length).

        Address ranges must be returned ordered by the virtual address.
        """
        _ = start
        return []

    def end(self):
        runs = list(self.get_available_addresses())
        if runs:
            return runs[-1][0] + runs[-1][2]

    def get_address_ranges(self, start=0, end=None):
        """Generates the address ranges which fall between start and end.

        Note that start and end are here specified in the virtual address
        space. More importantly this does not say anything about the pages in
        the physical address space - just because pages in the virtual address
        space are contiguous does not mean they are also contiguous in the
        physical address space.
        """
        if end is None:
            end = 0xfffffffffffff

        for voffset, poffset, length in self._get_address_ranges(
                start=start, end=end):
            # The entire range is below what is required - ignore it.
            if voffset + length < start:
                continue

            # The range starts after the address we care about - we are done.
            if voffset > end:
                return

            # Clip the bottom of the range to the start point, and the end of
            # the range to the end point.
            range_start = max(start, voffset)
            phys_range_start = poffset + range_start - voffset
            range_end = min(end, voffset + length)

            if range_end > range_start:
                yield range_start, phys_range_start, range_end - range_start

    def _get_address_ranges(self, start=0, end=None):
        """Generates merged address ranges from get_available_addresses()."""
        contiguous_voffset = 0
        contiguous_poffset = 0
        total_length = 0

        for (voffset, poffset, length) in self.get_available_addresses(
                start=start):

            if end and voffset > end:
                break

            # This can take some time as we enumerate all the address ranges.
            self.session.report_progress(
                "%(name)s: Merging Address Ranges %(offset)#x %(spinner)s",
                offset=voffset, name=self.name)

            # Try to join up adjacent pages as much as possible.
            if (voffset == contiguous_voffset + total_length and
                    poffset == contiguous_poffset + total_length):
                total_length += length

            else:
                if total_length > 0:
                    yield (contiguous_voffset, contiguous_poffset, total_length)

                # Reset the contiguous range.
                contiguous_voffset = voffset
                contiguous_poffset = poffset or 0
                total_length = length

        if total_length > 0:
            yield (contiguous_voffset, contiguous_poffset, total_length)

    def is_valid_address(self, _addr):
        """Tell us if the address is valid """
        return True

    def write(self, addr, buf):
        """Write to the address space, if writable.

        The default behavior is to delegate the write to the base address space.
        If an address space has no base then this function will throw an
        IOError. Address spaces that actually implement writing should override.

        Raises:
            IOError if there is no base address space. Subclasses may raise
                under additional circumstances.

        Arguments:
            addr: The address to write at, as understood by this AS (i.e.
                a virtual address for virtual address spaces, physical for
                physical).
            buf: The data to write - most commonly a basestring instance.

        Returns:
            Number of bytes written.
        """
        if not self.base:
            raise IOError("No base address space set on %r." % self)

        return self.base.write(self.vtop(addr), buf)

    def vtop(self, addr):
        """Return the physical address of this virtual address."""
        # For physical address spaces, this is a noop.
        return addr

    @classmethod
    def metadata(cls, name, default=None):
        """Obtain metadata about this address space."""
        return getattr(cls, "_%s__%s" % (cls.__name__, name), default)

    def __unicode__(self):
        return self.__class__.__name__

    def __str__(self):
        return utils.SmartStr(self)

    def __repr__(self):
        return "<%s @ %#x %s>" % (
            self.__class__.__name__, hash(self), self.name)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.base == other.base)


    def get_mapped_offset(self, filename, offset):
        """Implement this if we can map files into this address space."""


# This is a specialised AS for use internally - Its used to provide
# transparent support for a string buffer so types can be
# instantiated off the buffer.
class BufferAddressSpace(BaseAddressSpace):
    __abstract = True

    @property
    def writable(self):
        """Buffer AS is always writable, no matter what the session says."""
        return True

    def __init__(self, base_offset=0, data='', metadata=None, **kwargs):
        super(BufferAddressSpace, self).__init__(**kwargs)
        self.fname = "Buffer"
        self.data = data
        self.base_offset = base_offset
        self._metadata = metadata or {}

    def metadata(self, key):
        return self._metadata.get(key)

    def assign_buffer(self, data, base_offset=0):
        self.base_offset = base_offset
        self.data = data

    def is_valid_address(self, addr):
        return not (addr < self.base_offset or addr > self.base_offset +
                    len(self.data))

    def read(self, addr, length):
        offset = addr - self.base_offset
        data = self.data[offset: offset + length]
        return data + "\x00" * (length - len(data))

    def write(self, addr, data):
        if addr > len(self.data):
            raise ValueError(
                "Cannot write to offset %d of buffer with size %d.",
                addr, len(self.data))
        self.data = self.data[:addr] + data + self.data[addr + len(data):]
        return len(data)

    def get_available_addresses(self, start=None):
        yield (self.base_offset, self.base_offset, len(self.data))

    def get_buffer_offset(self, offset):
        """Returns the offset in self.data for the virtual offset."""
        return offset - self.base_offset

    def __repr__(self):
        return "<%s @ %#x %s [%#X-%#X]>" % (
            self.__class__.__name__, hash(self), self.name,
            self.base_offset, self.end())

    def __len__(self):
        return len(self.data)

    def end(self):
        """Return the end address of the buffer."""
        return self.base_offset + len(self.data)


class CachingAddressSpaceMixIn(object):
    # The size of chunks we cache. This should be large enough to make file
    # reads efficient.
    CHUNK_SIZE = 32 * 1024
    CACHE_SIZE = 1000

    def __init__(self, **kwargs):
        super(CachingAddressSpaceMixIn, self).__init__(**kwargs)
        self._cache = utils.FastStore(self.CACHE_SIZE)

    def read(self, addr, length):
        addr, length = int(addr), int(length)

        result = ""
        while length > 0:
            data = self.read_partial(addr, length)
            if not data:
                break

            result += data
            length -= len(data)
            addr += len(data)

        return result

    def read_partial(self, addr, length):
        if addr == None:
            return addr

        chunk_number = addr / self.CHUNK_SIZE
        chunk_offset = addr % self.CHUNK_SIZE

        # Do not cache large reads.
        if chunk_offset == 0 and length > self.CHUNK_SIZE:
            return super(CachingAddressSpaceMixIn, self).read(
                addr, length)

        available_length = min(length, self.CHUNK_SIZE - chunk_offset)

        try:
            data = self._cache.Get(chunk_number)
        except KeyError:
            # Just read the data from the real class.
            data = super(CachingAddressSpaceMixIn, self).read(
                chunk_number * self.CHUNK_SIZE, self.CHUNK_SIZE)

            self._cache.Put(chunk_number, data)

        return data[chunk_offset:chunk_offset + available_length]


class PagedReader(BaseAddressSpace):
    """An address space which reads in page size.

    This automatically takes care of splitting a large read into smaller reads.
    """
    PAGE_SIZE = 0x1000
    __abstract = True

    def _read_chunk(self, vaddr, length):
        """Read bytes from a virtual address.

        Args:
          vaddr: A virtual address to read from.
          length: The number of bytes to read.

        Returns:
          As many bytes as can be read within this page.
        """
        to_read = min(length, self.PAGE_SIZE - (vaddr % self.PAGE_SIZE))
        paddr = self.vtop(vaddr)
        if paddr is None:
            return "\x00" * to_read

        return self.base.read(paddr, to_read)

    def _write_chunk(self, vaddr, buf):
        to_write = min(len(buf), self.PAGE_SIZE - (vaddr % self.PAGE_SIZE))
        if not to_write:
            return 0

        paddr = self.vtop(vaddr)
        if not paddr:
            return 0

        return self.base.write(paddr, buf[:to_write])

    def write(self, addr, buf):
        available = len(buf)
        written = 0

        while available > written:
            chunk_len = self._write_chunk(addr + written, buf[written:])
            if not chunk_len:
                break
            written += chunk_len

        return written

    def read(self, addr, length):
        """Read 'length' bytes from the virtual address 'vaddr'."""
        if length > self.session.GetParameter("buffer_size"):
            raise IOError("Too much data to read.")

        addr, length = int(addr), int(length)

        result = ''

        while length > 0:
            buf = self._read_chunk(addr, length)
            if not buf:
                break

            result += buf
            addr += len(buf)
            length -= len(buf)

        return result

    def is_valid_address(self, addr):
        vaddr = self.vtop(addr)
        return vaddr != None and self.base.is_valid_address(vaddr)


class RunBasedAddressSpace(PagedReader):
    """An address space which uses a list of runs to specify a mapping."""

    # This is a list of (memory_offset, file_offset, length) tuples.
    runs = None
    __abstract = True

    def __init__(self, **kwargs):
        super(RunBasedAddressSpace, self).__init__(**kwargs)
        self.runs = utils.SortedCollection(key=lambda x: x[0])

        # Our get_available_addresses() refers to the base address space we
        # overlay on. RunBasedAddressSpace is simply a translation layer on top
        # of the base address space.
        if self.base is not None:
            self.phys_base = weakref.proxy(self.base)

    def _read_chunk(self, addr, length):
        """Read from addr as much as possible up to a length of length."""
        file_offset, available_length = self._get_available_buffer(
            addr, length)

        # Mapping not valid. We need to pad until the next run.
        if file_offset is None:
            return "\x00" * min(length, available_length)

        return self.base.read(file_offset, min(length, available_length))

    def _write_chunk(self, addr, buf):
        buflen = len(buf)
        file_offset, available_length = self._get_available_buffer(
            addr, buflen)

        length = min(buflen, available_length)

        # Silently skip unavailable runs.
        if file_offset is None or length == 0:
            return length

        return self.base.write(file_offset, buf[:length])

    def vtop(self, addr):
        file_offset, _ = self._get_available_buffer(addr, 1)
        return file_offset

    def _get_available_buffer(self, addr, length):
        """Resolves the address into the file offset.

        This function finds the run that contains this page and returns the file
        address where this page can be found.

        Returns:
          A tuple of (physical_offset, available_length). The
          physical_offset can be None to signify that the address is not
          valid. In this case the available_length signifies the number of
          bytes until the next available run.
        """
        addr = int(addr)
        try:
            virt_addr, file_address, file_length = self.runs.find_le(addr)
            available_length = file_length - (addr - virt_addr)
            physical_offset = addr - virt_addr + file_address

            if available_length > 0:
                return physical_offset, min(length, available_length)

        except ValueError:
            pass

        try:
            # Addr is outside any run, we need to find the next available
            # run and return the number of bytes we need to skip until then.
            virt_addr, _, _ = self.runs.find_ge(addr)

            return None, virt_addr - addr

        except ValueError:
            pass

        # A physical_offset of None means the address is not valid. If we get
        # here we dont have a next valid range.
        return None, 0xfffffffffffff

    def is_valid_address(self, addr):
        return self.vtop(addr) is not None

    def get_available_addresses(self, start=0):
        for run in self.runs:
            run_start, file_address, length = run[:3]
            if start > run_start + length:
                continue

            yield run_start, file_address, length

    def __eq__(self, other):
        return (super(RunBasedAddressSpace, self).__eq__(other) and
                self.runs == other.runs)


# TODO: Replace the RunBasedAddressSpace with this one since it is a super set.

class MultiRunBasedAddressSpace(PagedReader):
    """An address space which uses a list of runs to specify a mapping.

    This essentially delegates certain address ranges to other address spaces
    "mapped" into this address space.

    The runs are tuples of this form:

    (virtual_address, physical_address, length, address_space)

    - Virtual Address - An address in this address space's virtual address
      space.

    - Physical Address - An address in the delegate address space.

    - Length - The length of the mapped region.

    - Address space - the address space that should be read for this
      region. Note that the physical address above refers to addresses in this
      delegate address space.
    """

    # This is a list of (memory_offset, file_offset, length) tuples.
    runs = None
    __abstract = True

    def __init__(self, **kwargs):
        super(MultiRunBasedAddressSpace, self).__init__(**kwargs)
        self.runs = utils.SortedCollection(key=lambda x: x[0])
        self.phys_base = weakref.proxy(self)

    def add_run(self, virt_addr, file_address, file_len, address_space):
        self.runs.insert((virt_addr, file_address, file_len, address_space))

    def _read_chunk(self, addr, length):
        """Read from addr as much as possible up to a length of length."""
        try:
            virt_addr, file_address, file_len, _as = self.runs.find_le(addr)
            available_length = file_len - (addr - virt_addr)
            physical_offset = addr - virt_addr + file_address

            if available_length > 0:
                return _as.read(physical_offset, min(length, available_length))

        except ValueError:
            pass

        try:
            # Addr is outside any run, we need to find the next available
            # run and return the number of bytes we need to pad until then.
            virt_addr, _, _, _ = self.runs.find_ge(addr)

            return "\x00" * min(length, virt_addr - addr)

        except ValueError:
            # If we get here we dont have a next valid range.
            return "\x00" * length

    def vtop(self, addr):
        """Returns the physical address for this virtual address.

        Note that this does not mean much without also knowing the address space
        to read from. Maybe we need to change this method prototype?
        """
        try:
            virt_addr, file_address, file_len, _ = self.runs.find_le(addr)
            available_length = file_len - (addr - virt_addr)
            physical_offset = addr - virt_addr + file_address

            if available_length > 0:
                return physical_offset

        except ValueError:
            pass

    def is_valid_address(self, addr):
        return self.vtop(addr) is not None

    # FIXME: Deprecate this method in all address spaces in favor of
    # get_mappings() below.
    def get_available_addresses(self, start=0):
        for run_start, _, length, _ in self.runs:
            if start > run_start + length:
                continue

            # We can not allow the scanner to try to access the phys_base
            # directly because we have multiple phys_base and the scanner will
            # not know which to read. When the scanner uses get_mappings() this
            # should work again.
            yield run_start, run_start, length

    def get_mappings(self, start=0):
        """Returns the mappings."""
        for run_start, file_address, length, _as in self.runs:
            if start > run_start + length:
                continue

            yield run_start, file_address, length, _as



class Error(Exception):
    """Address space errors."""


class ASAssertionError(Error, IOError, AssertionError):
    """The address space failed to instantiate."""


class AddrSpaceError(Error):
    """Address Space Exception.

    This exception is raised when an AS decides to not be instantiated. It is
    used in the voting algorithm.
    """

    def __init__(self):
        self.reasons = []
        Error.__init__(self, "No suitable address space mapping found")

    def append_reason(self, driver, reason):
        self.reasons.append((driver, reason))

    def __str__(self):
        result = Error.__str__(self) + "\nTried to open image as:\n"
        for k, v in self.reasons:
            result += " {0}: {1}\n".format(k, v)

        return result



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Darwin Session collectors.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

from rekall.entities import definitions

from rekall.plugins.collectors.darwin import common
from rekall.plugins.collectors.darwin import zones


class DarwinTerminalUserInferor3000(common.DarwinEntityCollector):
    """Infers the relationship between usernames and UIDs using tty sessions."""

    outputs = ["User"]

    collect_args = dict(
        terminals=("Terminal/file matches (has component Permissions) and "
                   "Terminal/session"))

    complete_input = True

    def collect(self, hint, terminals):
        for terminal in terminals:
            owner = terminal["Terminal/file"]["Permissions/owner"]
            user = terminal["Terminal/session"]["Session/user"]
            # Now tell the manager that these two users are the same user.
            if owner and user:
                yield user.identity | owner.identity


class DarwinTTYZoneCollector(zones.DarwinZoneElementCollector):
    outputs = ["Struct/type=tty"]
    zone_name = "ttys"
    type_name = "tty"

    def validate_element(self, tty):
        return tty.t_lock == tty


class DarwinClistParser(common.DarwinEntityCollector):
    outputs = ["Buffer/purpose=terminal_input",
               "Buffer/purpose=terminal_output"]

    collect_args = dict(clists="Struct/type is 'clist'")

    def collect(self, hint, clists):
        for entity in clists:
            clist = entity["Struct/base"]
            yield [entity.identity,
                   definitions.Buffer(kind="ring",
                                      state="freed",
                                      contents=clist.recovered_contents,
                                      start=clist.c_cs,
                                      end=clist.c_ce,
                                      size=clist.c_cn)]


class DarwinTTYParser(common.DarwinEntityCollector):
    outputs = ["Terminal", "Struct/type=vnode", "Struct/type=clist",
               "Buffer/purpose=terminal_input",
               "Buffer/purpose=terminal_output"]
    collect_args = dict(ttys="Struct/type is 'tty'")

    def collect(self, hint, ttys):
        for entity in ttys:
            file_identity = None
            session_identity = None

            tty = entity["Struct/base"]
            session = tty.t_session.deref()
            vnode = session.s_ttyvp

            if session:
                session_identity = self.manager.identify({
                    "Struct/base": session})

            if vnode:
                # Look, it has a vnode!
                yield definitions.Struct(base=vnode,
                                         type="vnode")
                file_identity = self.manager.identify({
                    "Struct/base": vnode})

            # Yield just the stubs of the input and output ring buffers.
            # DarwinClistParser will grab these if it cares.
            yield [definitions.Struct(base=tty.t_rawq,
                                      type="clist"),
                   definitions.Buffer(purpose="terminal_input",
                                      context=entity.identity)]
            yield [definitions.Struct(base=tty.t_outq,
                                      type="clist"),
                   definitions.Buffer(purpose="terminal_output",
                                      context=entity.identity)]

            # Last, but not least, the Terminal itself.
            yield [entity.identity,
                   definitions.Terminal(
                       session=session_identity,
                       file=file_identity)]


class DarwinSessionParser(common.DarwinEntityCollector):
    """Collects session entities from the memory objects."""

    _name = "sessions"

    outputs = ["Session",
               "User",
               "Struct/type=tty",
               "Struct/type=proc"]

    collect_args = dict(sessions="Struct/type is 'session'")

    def collect(self, hint, sessions):
        for entity in sessions:
            session = entity["Struct/base"]

            # Have to sanitize the usernames to prevent issues when comparing
            # them later.
            username = str(session.s_login).replace("\x00", "")
            if username:
                user_identity = self.manager.identify({
                    "User/username": username})
                yield [user_identity,
                       definitions.User(
                           username=username)]
            else:
                user_identity = None

            sid = session.s_sid
            # Turns out, SID is not always unique. This is disabled as it is
            # not being currently used, and I need to investigate the causes
            # of duplicate sessions occurring on 10.10.
            # session_identity = self.manager.identify({
            #     "Session/sid": sid}) | entity.identity

            session_identity = entity.identity

            if session.s_ttyp:
                yield definitions.Struct(
                    base=session.s_ttyp,
                    type="tty")

            if session.s_leader and session.s_leader.validate():
                yield definitions.Struct(
                    base=session.s_leader.deref(),
                    type="proc")

            yield [session_identity,
                   definitions.Session(
                       user=user_identity,
                       sid=sid),
                   definitions.Named(
                       name="SID %d" % int(sid),
                       kind="Session")]


class DarwinSessionZoneCollector(zones.DarwinZoneElementCollector):
    """Collects sessions from the sessions allocation zone."""

    outputs = ["Struct/type=session"]
    zone_name = "session"
    type_name = "session"

    def validate_element(self, session):
        return session.s_count > 0 and session.s_leader.p_argc > 0


class DarwinSessionCollector(common.DarwinEntityCollector):
    """Collects sessions."""

    outputs = ["Struct/type=session"]

    def collect(self, hint):
        session_hash_table_size = self.profile.get_constant_object(
            "_sesshash", "unsigned long")

        # The hashtable is an array to session list heads.
        session_hash_table = self.profile.get_constant_object(
            "_sesshashtbl",
            target="Pointer",
            target_args=dict(
                target="Array",
                target_args=dict(
                    target="sesshashhead",
                    count=session_hash_table_size.v())))

        for sesshashhead in session_hash_table:
            for session in sesshashhead.lh_first.walk_list("s_hash.le_next"):
                yield definitions.Struct(
                    base=session,
                    type="session")



#!/usr/bin/env python2

# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Author: Michael Cohen scudette@google.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Tools for manipulating json output.

When decoding json output, the decoder may not have access to the original
image. Therefore we can not simply recreate the original encoded objects
(because they will need to read from the original image). We must therefore
create a standin for these objects which looks similar to the original but is
able to be used directly - i.e. without reading the original image.
"""

__author__ = "Michael Cohen <scudette@google.com>"

import json
from rekall import plugin
from rekall import testlib

from rekall.ui import json_renderer


class JSONParser(plugin.Command):
    """Renders a json rendering file, as produced by the JsonRenderer.

    The output of any plugin can be stored to a JSON file using:

    rekall -f img.dd --format json plugin_name --output test.json

    Then it can be rendered again using:

    rekall json_render test.json

    This plugin implements the proper decoding of the JSON encoded output.
    """

    name = "json_render"


    @classmethod
    def args(cls, parser):
        super(JSONParser, cls).args(parser)

        parser.add_argument("file", default=None,
                            help="The filename to parse.")

    def __init__(self, file=None, fd=None, **kwargs):
        super(JSONParser, self).__init__(**kwargs)

        # Make a json renderer to decode the json stream with.
        self.json_renderer = json_renderer.JsonRenderer(session=self.session)

        self.file = file
        self.fd = fd

    def RenderStatement(self, statement, renderer):
        """Renders one json decoded data command at a time."""
        command = statement[0]
        options = {}
        if command == "l":
            self.json_renderer.decoder.SetLexicon(statement[1])

        elif command == "m":
            renderer.section("Plugin %s" % statement[1]["plugin_name"])

        elif command == "s":
            renderer.section(
                **self.json_renderer.decoder.Decode(statement[1], options))

        elif command == "e":
            renderer.report_error(statement[1])

        elif command == "f":
            args = [self.json_renderer.decoder.Decode(x, options)
                    for x in statement[1:]]
            renderer.format(*args)

        elif command == "t":
            renderer.table_header(columns=statement[1])

        elif command == "r":
            row = [self.json_renderer.decoder.Decode(x, options)
                   for x in statement[1]]
            renderer.table_row(*row, **options)

    def render(self, renderer):
        """Renders the stored JSON file using the default renderer.

        To decode the json file we replay the statements into the renderer after
        decompressing them.
        """
        if self.file:
            self.fd = renderer.open(filename=self.file, mode="rb")

        if self.fd is None:
            raise ValueError("Need a filename or a file descriptor.")

        data = json.load(self.fd)

        for statement in data:
            self.RenderStatement(statement, renderer)


class TestJSONParser(testlib.SimpleTestCase):
    """Test the JSON renderer/parser."""
    PLUGIN = "json_render"

    PARAMETERS = dict(
        # The plugin to test json rendering with.
        commandline="pslist"
        )


    def BuildBaselineData(self, config_options):
        # We want to actually run the plugin first with JsonRenderer, then run
        # json_render on its json output - That will be the baseline.
        config_options["commandline"] = (
            "--format json -v --output %(tempdir)s_output.json " +
            config_options["commandline"])

        baseline = super(TestJSONParser, self).BuildBaselineData(config_options)

        output_file = self.temp_directory + "_output.json"
        config_options["commandline"] = "json_render %s" % output_file

        baseline = super(TestJSONParser, self).BuildBaselineData(config_options)
        return baseline



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Author: Michael Cohen scudette@google.com.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Decompression support for the LZNT1 compression algorithm.

Reference:
http://msdn.microsoft.com/en-us/library/jj665697.aspx
(2.5 LZNT1 Algorithm Details)

https://github.com/libyal/reviveit/
https://github.com/sleuthkit/sleuthkit/blob/develop/tsk/fs/ntfs.c
"""
import array
import cStringIO
import logging
import struct



def get_displacement(offset):
    """Calculate the displacement."""
    result = 0
    while offset >= 0x10:
        offset >>= 1
        result += 1

    return result


DISPLACEMENT_TABLE = array.array(
    'B', [get_displacement(x) for x in xrange(8192)])

COMPRESSED_MASK = 1 << 15
SIGNATURE_MASK = 3 << 12
SIZE_MASK = (1 << 12) - 1
TAG_MASKS = [(1 << i) for i in range(0, 8)]


def decompress_data(cdata, logger=None):
    """Decompresses the data."""

    if not logger:
        lznt1_logger = logging.getLogger("ntfs.lznt1")
    else:
        lznt1_logger = logger.getChild("lznt1")
    # Change to DEBUG to turn on module level debugging.
    lznt1_logger.setLevel(logging.ERROR)
    in_fd = cStringIO.StringIO(cdata)
    output_fd = cStringIO.StringIO()
    block_end = 0

    while in_fd.tell() < len(cdata):
        block_offset = in_fd.tell()
        uncompressed_chunk_offset = output_fd.tell()

        block_header = struct.unpack("<H", in_fd.read(2))[0]
        lznt1_logger.debug("Header %#x @ %#x", block_header, block_offset)
        if block_header & SIGNATURE_MASK != SIGNATURE_MASK:
            break

        size = (block_header & SIZE_MASK)
        lznt1_logger.debug("Block size %s", size + 3)

        block_end = block_offset + size + 3

        if block_header & COMPRESSED_MASK:
            while in_fd.tell() < block_end:
                header = ord(in_fd.read(1))
                lznt1_logger.debug("Tag %#x", header)
                for mask in TAG_MASKS:
                    if in_fd.tell() >= block_end:
                        break

                    if header & mask:
                        pointer = struct.unpack("<H", in_fd.read(2))[0]
                        displacement = DISPLACEMENT_TABLE[
                            output_fd.tell() - uncompressed_chunk_offset - 1]

                        symbol_offset = (pointer >> (12 - displacement)) + 1
                        symbol_length = (pointer & (0xFFF >> displacement)) + 3

                        output_fd.seek(-symbol_offset, 2)
                        data = output_fd.read(symbol_length)

                        # Pad the data to make it fit.
                        if 0 < len(data) < symbol_length:
                            data = data * (symbol_length / len(data) + 1)
                            data = data[:symbol_length]

                        output_fd.seek(0, 2)
                        lznt1_logger.debug(
                            "Wrote %s @ %s/%s: Phrase %s %s %x",
                            len(data), in_fd.tell(),
                            output_fd.tell(), symbol_length, symbol_offset,
                            pointer)

                        output_fd.write(data)

                    else:
                        data = in_fd.read(1)
                        lznt1_logger.debug("Symbol %#x", ord(data))
                        output_fd.write(data)

        else:
            # Block is not compressed
            data = in_fd.read(size + 1)
            output_fd.write(data)

    result = output_fd.getvalue()

    return result



from rekall import testlib

from rekall.entities.query import analyzer
from efilter import query as entity_query


class AnalyzerTest(testlib.RekallBaseUnitTestCase):
    def analyze(self, query):
        return entity_query.Query(query).run_engine("analyzer")

    def assertDepends(self, query, inclusions, exclusions=(), omissions=()):
        analysis = self.analyze(query)
        include = analysis.include
        exclude = analysis.exclude

        for dep in inclusions:
            dependency = analyzer.SimpleDependency.parse(dep)
            self.assertNotIn(dependency.inverted(), exclude)
            self.assertIn(dependency, include)

        for dep in exclusions:
            dependency = analyzer.SimpleDependency.parse(dep)
            self.assertNotIn(dependency, include)
            self.assertIn(dependency.inverted(), exclude)

        for dep in omissions:
            dependency = analyzer.SimpleDependency.parse(dep)
            self.assertNotIn(dependency, include)
            self.assertNotIn(dependency.inverted(), exclude)

    def testBasicDeps(self):
        query = "Process/command is 'foobar'"
        self.assertDepends(query, ["Process/command=foobar"], [], ["Process"])

    def testExclusions(self):
        query = "Struct/type is not 'socket'"
        self.assertDepends(query, ["Struct"],
                           ["Struct/type=socket"])

    def testMutuallyExclusive(self):
        query = ("Struct/type is 'socket' and "
                 "Struct/type is not 'socket'")
        self.assertDepends(query, ["Struct"], [],
                           ["Struct/type=socket"])



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__author__ = (
    "Michael Cohen <scudette@google.com>",
    "Adam Sindelar <adam.sindelar@gmail.com>")

from rekall.plugins.darwin import common


class DarwinArp(common.DarwinPlugin):
    """Show information about arp tables."""

    __name = "arp"

    def render(self, renderer):
        renderer.table_header(
            [("IP Addr", "ip_addr", "20"),
             ("MAC Addr", "mac", "18"),
             ("Interface", "interface", "9"),
             ("Sent", "sent", "8"),
             ("Recv", "recv", "8"),
             ("Time", "timestamp", "24"),
             ("Expires", "expires", "8"),
             ("Delta", "delta", "8")])

        arp_cache = self.profile.get_constant_object(
            "_llinfo_arp",
            target="Pointer",
            target_args=dict(target="llinfo_arp"))

        while arp_cache:
            entry = arp_cache.la_rt

            renderer.table_row(
                entry.source_ip,
                entry.dest_ip,
                entry.name,
                entry.sent,
                entry.rx,
                entry.base_calendartime,
                entry.rt_expire,
                entry.delta)

            arp_cache = arp_cache.la_le.le_next


class DarwinRoute(common.DarwinPlugin):
    """Show routing table."""

    __name = "route"

    RNF_ROOT = 2

    def rn_walk_tree(self, h):
        """Walks the radix tree starting from the header h.

        This function is taken from
        xnu-2422.1.72/bsd/net/radix.c: rn_walk_tree()

        Which is why it does not conform to the style guide.

        Note too that the darwin source code abuses C macros:

        #define rn_dupedkey     rn_u.rn_leaf.rn_Dupedkey
        #define rn_key          rn_u.rn_leaf.rn_Key
        #define rn_mask         rn_u.rn_leaf.rn_Mask
        #define rn_offset       rn_u.rn_node.rn_Off
        #define rn_left         rn_u.rn_node.rn_L
        #define rn_right        rn_u.rn_node.rn_R

        And then the original code does:
        rn = rn.rn_left

        So we replace these below.
        """
        rn = h.rnh_treetop

        seen = set()

        # First time through node, go left */
        while rn.rn_bit >= 0:
            rn = rn.rn_u.rn_node.rn_L

        while rn and rn not in seen:
            base = rn

            seen.add(rn)

            # If at right child go back up, otherwise, go right
            while (rn.rn_parent.rn_u.rn_node.rn_R == rn and
                   not rn.rn_flags & self.RNF_ROOT):
                rn = rn.rn_parent

            # Find the next *leaf* to start from
            rn = rn.rn_parent.rn_u.rn_node.rn_R
            while rn.rn_bit >= 0:
                rn = rn.rn_u.rn_node.rn_L

            next = rn

            # Process leaves
            while True:
                rn = base
                if not rn:
                    break

                base = rn.rn_u.rn_leaf.rn_Dupedkey
                if not rn.rn_flags & self.RNF_ROOT:
                    yield rn

            rn = next
            if rn.rn_flags & self.RNF_ROOT:
                return

    def render(self, renderer):
        renderer.table_header(
            [("Source IP", "source", "20"),
             ("Dest IP", "dest", "20"),
             ("Interface", "interface", "9"),
             ("Sent", "sent", "8"),
             ("Recv", "recv", "8"),
             ("Time", "timestamp", "24"),
             ("Expires", "expires", "8"),
             ("Delta", "delta", "8")])

        route_tables = self.profile.get_constant_object(
            "_rt_tables",
            target="Array",
            target_args=dict(
                count=32,
                target="Pointer",
                target_args=dict(
                    target="radix_node_head")))

        for node in self.rn_walk_tree(route_tables[2]):
            rentry = node.dereference_as("rtentry")

            renderer.table_row(
                rentry.source_ip,
                rentry.dest_ip,
                rentry.name,
                rentry.sent, rentry.rx,
                rentry.base_calendartime,
                rentry.rt_expire,
                rentry.delta)


class DarwinIPFilters(common.DarwinPlugin):
    """Check IP Filters for hooks."""

    __name = "ip_filters"

    def render(self, renderer):
        renderer.table_header([
            ("Context", "context", "10"),
            ("Filter", "filter", "16"),
            ("Handler", "handler", "[addrpad]"),
            ("Symbol", "symbol", "20")])

        resolver = self.session.address_resolver
        for list_name in ["_ipv4_filters", "_ipv6_filters"]:
            filter_list = self.profile.get_constant_object(
                list_name, target="ipfilter_list")

            for item in filter_list.tqh_first.walk_list("ipf_link.tqe_next"):
                filter = item.ipf_filter
                name = filter.name.deref()
                handler = filter.ipf_input.deref()
                renderer.table_row("INPUT", name, handler,
                                   resolver.format_address(handler))

                handler = filter.ipf_output.deref()
                renderer.table_row("OUTPUT", name, handler,
                                   resolver.format_address(handler))

                handler = filter.ipf_detach.deref()
                renderer.table_row("DETACH", name, handler,
                                   resolver.format_address(handler))



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""The module implements an OSX specific address resolution plugin."""

__author__ = "Michael Cohen <scudette@gmail.com>"
import re

from rekall import obj
from rekall import utils
from rekall.plugins.common import address_resolver
from rekall.plugins.darwin import common


class DarwinAddressResolver(address_resolver.AddressResolverMixin,
                            common.DarwinPlugin):
    """A Darwin specific address resolver plugin."""

    def __init__(self, **kwargs):
        super(DarwinAddressResolver, self).__init__(**kwargs)
        self.address_map = utils.SortedCollection(key=lambda x: x[0])

        # Delay initialization until we need it.
        self._initialized = False

    def _EnsureInitialized(self):
        if self._initialized:
            return

        modules = []
        modules.extend(self.session.plugins.lsmod().get_module_list())
        self.modules_by_name = {}
        for module in modules:
            self.address_map.insert((module.base, module))
            self.modules_by_name[module.name] = module

        self._initialized = True

    def _FindContainingModule(self, address):
        """Find the kernel module which contains the address."""
        self._EnsureInitialized()

        address = obj.Pointer.integer_to_address(address)
        try:
            _, module = self.address_map.find_le(address)
            if address < module.end:
                return module

        except ValueError:
            pass

        return obj.NoneObject("Unknown module")

    def FindContainingModule(self, address):
        """Finds the name of the module containing the specified address.

        In linux we currect only support kernel modules.

        Returns:
          A tuple of start address, end address, module name
          """
        self._EnsureInitialized()

        # The address may be in kernel space or user space. Currently we only
        # resolve addresses in kernel space.
        containing_module = self._FindContainingModule(address)

        return (containing_module.base, containing_module.size,
                containing_module.name)

    def LoadProfileForName(self, module):
        # Currently, on Darwin, we only support kernel symbols.  TODO: Develop a
        # MACHO export parser for exported kernel module symbols.
        if module == "__kernel__":
            return self.session.profile

        return obj.NoneObject("Module profiles are not supported yet.")

    def _format_address_from_profile(self, profile, address):
        nearest_offset, name = profile.get_nearest_constant_by_address(
            address)

        if name:
            difference = address - nearest_offset
            if difference == 0:
                return "%s!%s" % (profile.name, name)
            else:
                return "%s!%s+%#x" % (profile.name, name, difference)
        else:
            return "%s!+%#x" % (profile.name, address - profile.image_base)

    def get_nearest_constant_by_address(self, address):
        self._EnsureInitialized()

        address = obj.Pointer.integer_to_address(address)
        nearest_offset = 0
        full_name = module_name = symbol_name = ""
        profile = None

        # Find the containing module and see if we have a profile for it.
        containing_module = self._FindContainingModule(address)
        if containing_module:
            nearest_offset = containing_module.base
            full_name = module_name = containing_module.name

            # Try to load the module profile.
            profile = self.LoadProfileForName(module_name)
            if profile:
                offset, name = profile.get_nearest_constant_by_address(
                    address)

                # The profile's constant is closer than the module.
                if address - offset < address - nearest_offset:
                    nearest_offset = offset
                    symbol_name = name

            if symbol_name:
                full_name = "%s!%s" % (module_name, symbol_name)

        return nearest_offset, full_name

    def format_address(self, address, max_distance=0x1000):
        address = obj.Pointer.integer_to_address(address)

        # Try to locate the symbol below it.
        offset, name = self.get_nearest_constant_by_address(address)
        difference = address - offset

        if name:
            if difference == 0:
                return name

            # Ensure address falls within the current module.
            containing_module = self._FindContainingModule(address)
            if (containing_module and address < containing_module.end and
                0 < difference < max_distance):
                return "%s + %#x" % (
                    name, address - offset)

        return ""

    def search_symbol(self, pattern):
        # Currently we only allow searching in the same module.
        self._EnsureInitialized()
        result = []

        components = self._ParseAddress(pattern)
        module_name = components["module"]
        if module_name == None:
            raise RuntimeError(
                "Module name must be specified for symbol search.")

        profile = self.LoadProfileForName(module_name)

        # Match all symbols.
        symbol_regex = re.compile(components["symbol"].replace("*", ".*"))
        for constant in profile.constants:
            if symbol_regex.match(constant):
                result.append("%s!%s" % (module_name, constant))

        return result



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__author__ = "Michael Cohen <scudette@google.com>"

from rekall.plugins.darwin import common


class DarwinLsof(common.DarwinProcessFilter):
    """Lists open files, grouped by process that has the handle.

    A file is an overloaded term; this plugin will list files, directories,
    Unix sockets, pipes, shared memory regions and certain other kernel
    structures.
    """

    __name = "old_lsof"

    def lsof(self, proc_sort_key=None):
        """Get all open files (sockets, vnodes, etc.) for all processes.

        Args:
          proc_sort_key:
            A callable that takes proc and returns a key to sort by. If None
            (default) then no sorting will be done.

        Yields:
          Dict of proc, fd, flags and fileproc.
        """
        procs = self.filter_processes()
        if proc_sort_key:
            procs = sorted(procs, key=proc_sort_key)

        for proc in procs:
            for fd, fileproc, flags in proc.get_open_files():
                yield dict(proc=proc,
                           fd=fd,
                           flags=flags,
                           fileproc=fileproc)

    def render(self, renderer):
        renderer.table_header([("Command", "command", "16"),
                               ("PID", "pid", "8"),
                               ("UID", "uid", "8"),
                               ("FD", "fd", "10"),
                               ("Type", "type", "15"),
                               ("Name", "name", "40")])

        for open_file in self.lsof(proc_sort_key=lambda proc: proc.pid):
            renderer.table_row(open_file["proc"].p_comm,
                               open_file["proc"].pid,
                               open_file["proc"].p_uid,
                               open_file["fd"],
                               open_file["fileproc"].human_type,
                               open_file["fileproc"].human_name)



# Rekall Memory Forensics
# Copyright (C) 2008 Volatile Systems
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
This module implements the fast connection scanning

@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""

# pylint: disable=protected-access

from rekall.plugins.overlays.windows import tcpip_vtypes
from rekall.plugins.windows import common


class PoolScanConnFast(common.PoolScanner):
    checks = [('PoolTagCheck', dict(tag="TCPT")),
              ('CheckPoolSize', dict(condition=lambda x: x >= 0x198)),
              ('CheckPoolType', dict(non_paged=True, free=True)),
              ('CheckPoolIndex', dict(value=0)),
              ]

class ConnScan(tcpip_vtypes.TcpipPluginMixin,
               common.PoolScannerPlugin):
    """ Scan Physical memory for _TCPT_OBJECT objects (tcp connections)
    """

    __name = "connscan"

    @classmethod
    def is_active(cls, session):
        # These only work for XP.
        return (super(ConnScan, cls).is_active(session) and
                session.profile.metadata("major") == 5)

    def generate_hits(self):
        """Search the physical address space for _TCPT_OBJECTs.

        Yields:
          _TCPT_OBJECT instantiated on the physical address space.
        """
        scanner = PoolScanConnFast(
            session=self.session, profile=self.tcpip_profile,
            address_space=self.address_space)

        for pool_obj in scanner.scan():
            # The struct is allocated out of the pool (i.e. its not an object).
            yield self.tcpip_profile._TCPT_OBJECT(
                vm=self.address_space,
                offset=pool_obj.obj_offset + pool_obj.obj_size)

    def render(self, renderer):
        renderer.table_header([("Offset(P)", "offset_p", "[addrpad]"),
                               ("Local Address", "local_net_address", "<25"),
                               ("Remote Address", "remote_net_address", "<25"),
                               ("Pid", "pid", ">10")])

        ## We make a new scanner
        for tcp_obj in self.generate_hits():
            local = "{0}:{1}".format(tcp_obj.LocalIpAddress,
                                     tcp_obj.LocalPort)

            remote = "{0}:{1}".format(tcp_obj.RemoteIpAddress,
                                      tcp_obj.RemotePort)

            renderer.table_row(tcp_obj.obj_offset, local, remote, tcp_obj.Pid)



# -*- coding: utf-8 -*-

# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This module implements base object renderers."""

from rekall import utils

from rekall.ui import text


class BaseObjectTextRenderer(text.TextObjectRenderer):
    renders_type = "BaseObject"

    def render_address(self, target, **options):
        return text.Cell(
            self.format_address(target.obj_offset, **options)
        )

    def render_full(self, target, **options):
        result = text.Cell(unicode(target.v()), **options)
        return result

    def render_value(self, target, **_):
        return text.Cell(unicode(target.v()))


class StringTextRenderer(BaseObjectTextRenderer):
    renders_type = "String"

    def render_full(self, target, **_):
        return text.Cell(
            utils.SmartUnicode(target).split("\x00")[0] or u"")

    render_value = render_full
    render_compact = render_full


class NoneObjectTextRenderer(BaseObjectTextRenderer):
    """NoneObjects will be rendered with a single dash '-'."""
    renders_type = "NoneObject"

    def render_row(self, target, **_):
        return text.Cell("-")


class NoneTextRenderer(NoneObjectTextRenderer):
    renders_type = "NoneType"


class UnixTimestampObjectRenderer(BaseObjectTextRenderer):
    renders_type = "UnixTimeStamp"
    renderers = ["TextRenderer"]

    def render_row(self, target, details=False, **options):
        if details:
            return text.Cell(repr(target))

        if target:
            dt = target.as_datetime()
            if dt:
                return text.Cell(target.display_datetime(dt))

        return text.Cell("-")


class PythonBoolTextRenderer(text.TextObjectRenderer):
    renders_type = "bool"

    def render_full(self, target, **_):
        color = "GREEN" if target else "RED"
        return text.Cell(
            value=unicode(target),
            highlights=[(0, -1, color, None)])

    render_value = render_full
    render_compact = render_full


class PythonSetRenderer(text.TextObjectRenderer):
    renders_type = "frozenset"

    def render_full(self, target, **_):
        return text.Cell(repr(target))

    def render_compact(self, target, **_):
        return text.Cell(
            "{%s}" % ", ".join([repr(x) for x in target]))


class NativeTypeTextRenderer(BaseObjectTextRenderer):
    renders_type = "NativeType"

    def render_address(self, target, width=None, **options):
        return text.Cell(
            self.format_address(target.v(), **options),
            width=width)


class BaseBoolTextRenderer(PythonBoolTextRenderer):
    renders_type = "Bool"

    def render_row(self, target, **kwargs):
        return super(BaseBoolTextRenderer, self).render_row(bool(target),
                                                            **kwargs)


class FlagsTextRenderer(BaseObjectTextRenderer):
    renders_type = "Flags"

    def render_full(self, target, **_):
        flags = []
        value = target.v()
        for k, v in sorted(target.maskmap.items()):
            if value & v:
                flags.append(k)

        return text.Cell(u', '.join(flags))

    def render_value(self, target, **_):
        return text.Cell(unicode(self.v()))

    def render_compact(self, target, **_):
        lines = self.render_full(target).lines
        if not lines:
            return text.Cell("-")

        elided = lines[0]
        if len(elided) > 40:
            elided = elided[:39] + u"…"

        return text.Cell(elided)


class EnumerationTextRenderer(BaseObjectTextRenderer):
    renders_type = "Enumeration"

    def render_full(self, target, **_):
        value = target.v()
        name = target.choices.get(utils.SmartStr(value), target.default) or (
            u"UNKNOWN (%s)" % utils.SmartUnicode(value))

        return text.Cell(name)

    render_compact = render_full


class DatetimeTextRenderer(text.TextObjectRenderer):
    renders_type = "datetime"

    def render_row(self, target, **_):
        return text.Cell(target.strftime("%Y-%m-%d %H:%M:%S%z"))


class PointerTextRenderer(NativeTypeTextRenderer):
    renders_type = "Pointer"

    def render_value(self, *args, **kwargs):
        return self.render_address(*args, **kwargs)

    def render_full(self, target, **_):
        target_obj = target.deref()
        renderer = target_obj.get_text_renderer()
        return renderer.render_full(target_obj)

    def render_compact(self, target, **options):
        return text.Cell(
            "(%s *) %s" % (
                target.target,
                self.format_address(target.v(), **options))
        )


class ListRenderer(text.TextObjectRenderer):
    """Renders a long as an address."""
    renders_type = ("list", "tuple")

    def render_row(self, target, **options):
        result = []
        for item in target:
            object_renderer = self.ForTarget(item, self.renderer)(
                session=self.session, renderer=self.renderer)
            result.append(object_renderer.render_row(item, **options))
            result.append(text.Cell(","))

        if result:
            result.pop(-1)

        return text.JoinedCell(*result)


class VoidTextRenderer(PointerTextRenderer):
    renders_type = "Void"

    def render_full(self, target, **options):
        return text.Cell(
            "(void *) %s" % self.format_address(target.v(), **options))

    render_compact = render_full


class FunctionTextRenderer(BaseObjectTextRenderer):
    renders_type = "Function"

    def render_full(self, target, **_):
        if target.mode == "AMD64":
            format_string = "%0#14x  %s"
        else:
            format_string = "%0#10x  %s"

        result = []
        for offset, _, instruction in target.Disassemble():
            result.append(format_string % (offset, instruction))

        return text.Cell("\n".join(result))

    def render_compact(self, target, **options):
        return text.Cell(self.format_address(target.obj_offset, **options))

    render_value = render_compact


class StructTextRenderer(text.TextObjectRenderer):
    renders_type = "Struct"
    DEFAULT_STYLE = "compact"

    def render_full(self, target, **_):
        result = repr(target) + "\n"
        width_name = 0

        fields = []
        # Print all the fields sorted by offset within the struct.
        for k in target.members:
            width_name = max(width_name, len(k))
            obj = getattr(target, k)
            if obj == None:
                obj = target.m(k)

            fields.append(
                (getattr(obj, "obj_offset", target.obj_offset) -
                 target.obj_offset, k, utils.SmartUnicode(repr(obj))))

        fields.sort()

        result = result + u"\n".join(
            [u"  0x%02X %s%s %s" % (offset, k, " " * (width_name - len(k)), v)
             for offset, k, v in fields]) + "\n"

        return text.Cell(result)

    def render_compact(self, target, **_):
        return text.Cell(repr(target))



# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import logging


from rekall import obj
from rekall.plugins.windows.gui import constants
from rekall.plugins.windows.gui import win32k_core

from rekall.plugins.windows.gui.vtypes import win7_sp0_x64_vtypes_gui
from rekall.plugins.windows.gui.vtypes import win7_sp1_x64_vtypes_gui
from rekall.plugins.windows.gui.vtypes import win7_sp0_x86_vtypes_gui
from rekall.plugins.windows.gui.vtypes import win7_sp1_x86_vtypes_gui


class _MM_SESSION_SPACE(win32k_core._MM_SESSION_SPACE):
    """A class for session spaces on Windows 7"""

    def find_shared_info(self):
        """The way we find win32k!gSharedInfo on Windows 7
        is different than before. For each DWORD in the
        win32k.sys module's .data section (DWORD-aligned)
        we check if its the HeEntrySize member of a possible
        tagSHAREDINFO structure. This should equal the size
        of a _HANDLEENTRY.

        The HeEntrySize member didn't exist before Windows 7
        thus the need for separate methods."""
        handle_table_size = self.obj_profile.get_obj_size("_HANDLEENTRY")

        handle_entry_offset = self.obj_profile.get_obj_offset(
            "tagSHAREDINFO", "HeEntrySize")

        import pdb; pdb.set_trace()

        for chunk in self._section_chunks(".data"):

            if chunk != handle_table_size:
                continue

            shared_info = self.obj_profile.tagSHAREDINFO(
                offset = chunk.obj_offset - handle_entry_offset,
                vm = self.obj_vm)

            if shared_info.is_valid():
                return shared_info

        return obj.NoneObject("Cannot find win32k!gSharedInfo")


class tagSHAREDINFO(win32k_core.tagSHAREDINFO):
    """A class for shared info blocks on Windows 7"""

    def is_valid(self):
        """Sanity checks for tagSHAREDINFO"""

        if not super(tagSHAREDINFO, self).is_valid():
            return False

        if self.ulSharedDelta != 0:
            return False

        if not self.psi.is_valid():
            return False

        return (self.psi.cbHandleTable / self.HeEntrySize ==
                self.psi.cHandleEntries)


class Win32GUIWin7(obj.ProfileModification):
    """Installs the win7 specific profiles for the GUI modules."""

    @classmethod
    def modify(cls, profile):
        version = ".".join(profile.metadatas('major', 'minor'))
        build = profile.metadata("build", 7601)
        architecture = profile.metadata("arch")

        if architecture == "AMD64":
            # http://doxygen.reactos.org/d5/dd0/timer_8h_source.html#l00019
            profile.add_overlay({
                    'tagTIMER' : [None, {
                            'head' : [0x00, ['_HEAD']],
                            'ListEntry' : [0x18, ['_LIST_ENTRY']],
                            'spwnd' : [0x28, ['pointer', ['tagWND']]],
                            'pti' : [0x30, ['pointer', ['tagTHREADINFO']]],
                            'nID' : [0x38, ['unsigned short']],
                            'cmsCountdown' : [0x40, ['unsigned int']],
                            'cmsRate' : [0x44, ['unsigned int']],
                            'flags' : [0x48, ['Flags', dict(
                                        bitmap=constants.TIMER_FLAGS)]],
                            'pfn' : [0x50, ['pointer', ['void']]],
                            }]})

            if build == 7600:
                profile.add_overlay(win7_sp0_x64_vtypes_gui.win32k_types)
            elif build == 7601:
                profile.add_overlay(win7_sp1_x64_vtypes_gui.win32k_types)
            else:
                logging.warning("Unsupported version %s, will use win7sp1",
                                version)

                profile.add_overlay(win7_sp1_x64_vtypes_gui.win32k_types)

        elif architecture == "I386":
            # http://doxygen.reactos.org/d5/dd0/timer_8h_source.html#l00019
            profile.vtypes.update({
                    'tagTIMER' : [None, {
                            'ListEntry' : [0xc, ['_LIST_ENTRY']],
                            'pti' : [0x18, ['pointer', ['tagTHREADINFO']]],
                            'spwnd' : [0x14, ['pointer', ['tagWND']]], #??
                            'nID' : [0x1C, ['unsigned short']],
                            'cmsCountdown' : [0x20, ['unsigned int']],
                            'cmsRate' : [0x24, ['unsigned int']],
                            'flags' : [0x28, ['Flags', dict(
                                        bitmap=constants.TIMER_FLAGS)
                                              ]],
                            'pfn' : [0x2C, ['pointer', ['void']]],
                            }]})

            if build == 7600:
                profile.add_overlay(win7_sp0_x86_vtypes_gui.win32k_types)

            elif build == 7601:
                profile.add_overlay(win7_sp1_x86_vtypes_gui.win32k_types)

            else:
                logging.warning("Unsupported version %s, will use win7sp1",
                                version)

                profile.add_overlay(win7_sp1_x86_vtypes_gui.win32k_types)

        profile.add_overlay({
                'tagHOOK': [None, {
                        'flags': [None, ['Flags', dict(
                                    bitmap=constants.HOOK_FLAGS
                                    )]],
                        }],
                '_HANDLEENTRY': [None, {
                        'bType': [None, ['Enumeration', dict(
                                    target='unsigned char',
                                    choices=constants.HANDLE_TYPE_ENUM_SEVEN
                                    )]],
                        }],
                'tagWINDOWSTATION' : [None, {
                        'pClipBase' : [None, ['Pointer', dict(
                                    target="Array",
                                    target_args=dict(
                                        count=lambda x : x.cNumClipFormats,
                                        target='tagCLIP'
                                        ),
                                    )]],
                        }],
                'tagCLIP': [None, {
                        'fmt' : [None, ['Enumeration', dict(
                                    target='unsigned long',
                                    choices=constants.CLIPBOARD_FORMAT_ENUM
                                    )]],
                        }]})

        profile.add_classes({
            '_MM_SESSION_SPACE': _MM_SESSION_SPACE,
            'tagSHAREDINFO': tagSHAREDINFO,
            })

        return profile



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Base classes for all tests.

Rekall Memory Forensics tests have these goals:

- Detect regression bugs with previous versions of Rekall Memory Forensics.

- Once differences are detected, make it easier to understand why the
  differences arise.

- Sometime the differences are acceptable, in that case, there need to be a way
  to declare the allowed differences in the tests.

- What makes this more complicated is that the same test needs to be applied to
  multiple images in a consistent way without necessarily making any code
  changes to the test itself.


Solution Outline
----------------

A baseline of running each test is written by the test suite itself. The user
can write a baseline for all the modules by issuing the make_suite binary. The
baseline for each module will generate a number of files in a given directory.

When test is run, the baseline files are loaded and copared with present output
in a specific way.

"""
import hashlib
import logging
import subprocess
import pdb
import os
import sys
import unittest

from rekall import config
from rekall import plugin
from rekall import registry
from rekall import session as rekall_session


class RekallBaseUnitTestCase(unittest.TestCase):
    """Base class for all rekall unit tests."""
    __metaclass__ = registry.MetaclassRegistry
    __abstract = True

    # The parameters to run this test with. These parameters are written to the
    # config file when creating a new blank template. Users can edit the config
    # file to influence how the test is run.

    PARAMETERS = {
        # This is the command line which is used to run the test.
        "commandline": "",
    }

    PLUGIN = None

    disabled = False

    @classmethod
    def is_active(cls, _):
        return True

    @classmethod
    def CommandName(cls):
        if cls.PLUGIN:
            return cls.PLUGIN

        name = cls.PARAMETERS.get("commandline", "").split()
        if name:
            return name[0]

    def __init__(self, method_name="__init__", baseline=None, current=None,
                 debug=False, temp_directory=None, config_options=None):

        self.baseline = baseline
        self.config_options = config_options
        self.current = current
        self.debug = debug
        self.temp_directory = temp_directory
        super(RekallBaseUnitTestCase, self).__init__(method_name)

    def LaunchExecutable(self, config_options):
        """Launches the rekall executable with the config specified.

        Returns:
          A baseline data structure which contains meta data from running
          rekall over the test case.
        """
        tmp_filename = os.path.join(self.temp_directory,
                                    "." + self.__class__.__name__)
        error_filename = tmp_filename + ".stderr"

        baseline_commandline = config_options.get("commandline")

        # Nothing to do here.
        if not baseline_commandline:
            return {}

        # The command line is specified in the test's PARAMETERS dict.
        try:
            baseline_commandline = baseline_commandline % config_options
        except KeyError as e:
            logging.critical(
                "Test %s requires parameter %s to be set in config file. (%s)",
                config_options["test_class"], e, baseline_commandline)
            return {}

        if baseline_commandline:
            baseline_commandline = "- %s" % baseline_commandline
            for k, v in config_options.items():
                # prepend all global options to the command line.
                if k.startswith("-"):
                    # This is a boolean flag.
                    if v is True:
                        baseline_commandline = "%s %s" % (
                            k, baseline_commandline)

                    elif isinstance(v, list):
                        baseline_commandline = "%s %s %s" % (
                            k, " ".join("'%s'" % x for x in v),
                            baseline_commandline)

                    else:
                        baseline_commandline = "%s '%s' %s" % (
                            k, v, baseline_commandline)

            cmdline = (config_options["executable"] + " -v " +
                       baseline_commandline)
            logging.debug("%s: Launching %s", self.__class__.__name__, cmdline)

            config_options["executed_command"] = cmdline

            with open(tmp_filename, "wb") as output_fd:
                with open(error_filename, "wb") as error_fd:
                    pipe = subprocess.Popen(cmdline, shell=True,
                                            stdout=output_fd, stderr=error_fd)

                    pipe.wait()

                    # Done running the command, now prepare the json baseline
                    # file.
                    output_fd.flush()
                    error_fd.flush()

                    output = open(tmp_filename).read(10 * 1024 * 1024)
                    output = output.decode("utf8", "ignore")

                    error = open(error_filename).read(10 * 1024 * 1024)
                    error = error.decode("utf8", "ignore")

                    baseline_data = dict(output=output.splitlines(),
                                         logging=error.splitlines(),
                                         return_code=pipe.returncode)

                    return baseline_data

        else:
            # No valid command line - this baseline is aborted.
            config_options["aborted"] = True

            return {}

    def BuildBaselineData(self, config_options):
        return self.LaunchExecutable(config_options)

    def MakeUserSession(self, config_options=None):
        if config_options is None:
            config_options = self.config_options

        user_session = rekall_session.InteractiveSession()
        with user_session.state as state:
            config.MergeConfigOptions(state)
            for k, v in config_options.items():
                if k.startswith("--"):
                    state.Set(k[2:], v)

        return user_session

    def assertListEqual(self, a, b, msg=None):
        a = list(a)
        b = list(b)
        self.assertEqual(len(a), len(b))

        for x, y in zip(a, b):
            self.assertEqual(x, y)

    def __unicode__(self):
        return "%s %s" % (self.__class__.__name__, self._testMethodName)

    def run(self, result=None):
        if result is None:
            result = self.defaultTestResult()

        result.startTest(self)
        testMethod = getattr(self, self._testMethodName)
        try:
            try:
                self.setUp()
            except KeyboardInterrupt:
                raise
            except Exception:
                if self.debug:
                    pdb.post_mortem()

                result.addError(self, sys.exc_info())
                return

            ok = False
            try:
                testMethod()
                ok = True
            except self.failureException:
                if self.debug:
                    pdb.post_mortem()

                result.addFailure(self, sys.exc_info())
            except KeyboardInterrupt:
                raise
            except Exception:
                if self.debug:
                    pdb.post_mortem()

                result.addError(self, sys.exc_info())

            try:
                self.tearDown()
            except KeyboardInterrupt:
                raise
            except Exception:
                if self.debug:
                    pdb.post_mortem()

                result.addError(self, sys.exc_info())
                ok = False
            if ok:
                result.addSuccess(self)
        finally:
            result.stopTest(self)


class SimpleTestCase(RekallBaseUnitTestCase):
    """A simple test which just compares with the baseline output."""

    __abstract = True

    @classmethod
    def is_active(cls, session):
        delegate_plugin = (
            plugin.Command.ImplementationByClass(cls.PLUGIN) or
            getattr(session.plugins, cls.CommandName() or "", None))

        if delegate_plugin:
            return delegate_plugin.is_active(session)

    def testCase(self):
        previous = self.baseline['output']
        current = self.current['output']

        # Compare the entire table
        self.assertEqual(previous, current)


class SortedComparison(SimpleTestCase):

    __abstract = True

    def testCase(self):
        previous = sorted(self.baseline['output'])
        current = sorted(self.current['output'])

        # Compare the entire table
        self.assertListEqual(previous, current)


class DisabledTest(RekallBaseUnitTestCase):
    """Disable a test."""
    disabled = True

    @classmethod
    def is_active(cls, _):
        return False


class HashChecker(SimpleTestCase):
    """A test comparing the hashes of all the files dumped in the tempdir."""

    def BuildBaselineData(self, config_options):
        """We need to calculate the hash of the image we produce."""
        baseline = super(HashChecker, self).BuildBaselineData(config_options)
        baseline['hashes'] = {}
        for filename in os.listdir(self.temp_directory):
            if not filename.startswith("."):
                with open(os.path.join(self.temp_directory, filename)) as fd:
                    md5 = hashlib.md5()
                    while 1:
                        data = fd.read(1024 * 1024)
                        if not data:
                            break

                        md5.update(data)

                    baseline['hashes'][filename] = md5.hexdigest()

        return baseline

    def testCase(self):
        self.assertEqual(self.baseline['hashes'], self.current['hashes'])



from rekall import testlib

from rekall.entities import entity


class EntityTest(testlib.RekallBaseUnitTestCase):
    def testFromAsDict(self):
        data = {"Process/pid": 1,
                "Process/command": "init"}

        x = entity.Entity.fromdict(data)
        self.assertEqual(x.asdict(), data)



# Rekall Memory Forensics
# Copyright (C) 2008 Volatile Systems
# Copyright (c) 2008 Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       AAron Walters and Brendan Dolan-Gavitt
@license:      GNU General Public License 2.0 or later
@contact:      awalters@volatilesystems.com,bdolangavitt@wesleyan.edu
@organization: Volatile Systems
"""

from rekall import utils

from rekall.plugins.windows.registry import lsasecrets
from rekall.plugins.windows.registry import hashdump
from rekall.plugins.windows import common
from rekall.plugins.windows.registry import registry


class LSADump(common.WindowsCommandPlugin):
    """Dump (decrypted) LSA secrets from the registry"""
    # Declare meta information associated with this plugin

    __name = "lsadump"

    meta_info = {}
    meta_info['author'] = 'Brendan Dolan-Gavitt'
    meta_info['copyright'] = 'Copyright (c) 2007,2008 Brendan Dolan-Gavitt'
    meta_info['contact'] = 'bdolangavitt@wesleyan.edu'
    meta_info['license'] = 'GNU General Public License 2.0 or later'
    meta_info['url'] = 'http://moyix.blogspot.com/'
    meta_info['os'] = 'WIN_32_XP_SP2'
    meta_info['version'] = '1.0'

    @classmethod
    def is_active(cls, session):
        """We are only active for windows xp."""
        if session and session.profile:
            return session.profile.metadata("major") == 5

    def __init__(self, sys_offset=None, security_offset=None, **kwargs):
        """Dump (decrypted) LSA secrets from the registry.

        Args:
           sys_offset: The hive virtual offset to the system hive.
           security_offset: The hive virtual offset to the security hive.
        """
        super(LSADump, self).__init__(**kwargs)
        self.sys_offset = sys_offset
        self.security_offset = security_offset
        self.profile = registry.RekallRegisteryImplementation(self.profile)

    def calculate(self):
        sys_hive = registry.RegistryHive(
            profile=self.profile, hive_offset=self.sys_offset,
            kernel_address_space=self.kernel_address_space)

        security_hive = registry.RegistryHive(
            profile=self.profile, hive_offset=self.security_offset,
            kernel_address_space=self.kernel_address_space)

        return lsasecrets.get_secrets(sys_hive, security_hive)

    def render(self, outfd):
        for k, v in self.calculate():
            outfd.write(k + "\n")
            utils.WriteHexdump(outfd, v)
            outfd.write("\n")


class HashDump(LSADump):
    """Dumps passwords hashes (LM/NTLM) from memory"""

    __name = "hashdump"

    def __init__(self, sys_offset=None, sam_offset=None, **kwargs):
        """Dump (decrypted) LSA secrets from the registry.

        Args:
           sys_offset: The hive virtual offset to the system hive.
           sam_offset: The hive virtual offset to the sam hive.
        """
        super(HashDump, self).__init__(**kwargs)
        self.sys_offset = sys_offset
        self.sam_offset = sam_offset
        self.profile = registry.RekallRegisteryImplementation(self.profile)

    def calculate(self):
        sys_registry = registry.RegistryHive(
            profile=self.profile, hive_offset=self.sys_offset,
            kernel_address_space=self.kernel_address_space)

        sam_registry = registry.RegistryHive(
            profile=self.profile, hive_offset=self.sam_offset,
            kernel_address_space=self.kernel_address_space)

        return hashdump.dump_hashes(sys_registry, sam_registry)

    def render(self, outfd):
        for d in self.calculate():
            outfd.write(d + "\n")



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

# References:
# http://www.codemachine.com/article_kernelstruct.html#MMPFN
# http://www.reactos.org/wiki/Techwiki:Memory_management_in_the_Windows_XP_kernel#MmPfnDatabase

# pylint: disable=protected-access

from rekall import testlib
from rekall import obj
from rekall import plugin
from rekall.plugins import core
from rekall.plugins.windows import common
from rekall.plugins.overlays import basic

class ValueEnumeration(basic.Enumeration):
    """An enumeration which receives its value from a callable."""

    def __init__(self, value=None, parent=None, **kwargs):
        super(ValueEnumeration, self).__init__(parent=parent, **kwargs)
        if callable(value):
            value = value(parent)

        self.value = value

    def v(self, vm=None):
        return self.value


class PFNModification(obj.ProfileModification):
    """Installs types specific to the PFN database."""

    @classmethod
    def modify(cls, profile):
        # Some shortcuts to the most important information.
        profile.add_overlay({
            '_MMPTE': [None, {
                'Valid': lambda x: x.u.Hard.Valid,
                'PFN': lambda x: x.u.Hard.PageFrameNumber,
                }],
            '_MMPFN': [None, {
                "Type": [0, ["ValueEnumeration", dict(
                    value=lambda x: x.u3.e1.PageLocation,
                    choices={
                        0: 'ZeroedPageList',
                        1: 'FreePageList',
                        2: 'StandbyPageList',
                        3: 'ModifiedPageList',
                        4: 'ModifiedNoWritePageList',
                        5: 'BadPageList',
                        6: 'ActiveAndValid',
                        7: 'TransitionPage'
                        }
                    )]],
                }],
            '_KDDEBUGGER_DATA64': [None, {
                # This is the pointer to the PFN database.
                'MmPfnDatabase': [None, ['Pointer', dict(
                    target="Pointer",
                    target_args=dict(
                        target="Array",
                        target_args=dict(target="_MMPFN"),
                        ))]],
                }],
            })
        profile.add_classes({
            "ValueEnumeration": ValueEnumeration,
            })


class VtoP(core.VtoPMixin, common.WinProcessFilter):
    """Prints information about the virtual to physical translation."""


class PFNInfo(common.WindowsCommandPlugin):
    """Prints information about an address from the PFN database."""

    __name = "pfn"

    # Size of page.
    PAGE_SIZE = 0x1000
    PAGE_BITS = 12

    @classmethod
    def args(cls, parser):
        super(PFNInfo, cls).args(parser)
        parser.add_argument("pfn", type="IntParser",
                            help="The PFN to examine.")

    def __init__(self, pfn=None, physical_address=None, **kwargs):
        """Prints information about the physical PFN entry.

        Args:
          pfn: A page file number to display.
          physical_address: The physical address to print information about.
        """
        super(PFNInfo, self).__init__(**kwargs)

        self.profile = PFNModification(self.profile)

        # A reference to the pfn database.
        self.pfn_database = self.profile.get_constant_object(
            "MmPfnDatabase",
            target="Pointer",
            target_args=dict(
                target="Array",
                target_args=dict(
                    target="_MMPFN",
                    )
                )
            )

        self.pfn = pfn
        self.physical_address = physical_address

    def pfn_record(self, pfn=None, physical_address=None):
        """Returns the pfn record for a pfn or a virtual address."""
        if physical_address is not None:
            pfn = int(physical_address) / self.PAGE_SIZE

        if pfn is None:
            raise RuntimeError("PFN not provided.")

        # Return the pfn record.
        return self.pfn_database.deref()[pfn]

    def render(self, renderer):
        pfn = self.pfn
        if pfn is None:
            raise plugin.PluginError("PFN not provided.")

        if self.physical_address is not None:
            pfn = int(self.physical_address) / self.PAGE_SIZE

        pfn_obj = self.pfn_record(pfn)

        renderer.format("    PFN {0:style=address} at "
                        "kernel address {1:addrpad}\n",
                        pfn, pfn_obj.obj_offset)

        # The flags we are going to print.
        flags = {"M": "Modified",
                 "P": "ParityError",
                 "R": "ReadInProgress",
                 "W": "WriteInProgress"}

        short_flags_string = "".join(
            [k for k, v in flags.items() if pfn_obj.u3.e1.m(v) == 0])

        long_flags_string = " ".join(
            [v for k, v in flags.items() if pfn_obj.u3.e1.m(v) == 0])

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_physical_address = ((containing_page << self.PAGE_BITS) |
                                (int(pfn_obj.PteAddress) & 0xFFF))

        renderer.format("""    flink  {0:addr}  blink / share count {1:addr}
    pteaddress (VAS) {2:addrpad}  (Phys AS) {3:addr}
    reference count {4:addr}   color {5}
    containing page        {6:addr}  {7}     {8}
    {9}
    """, pfn_obj.u1.Flink, pfn_obj.u2.Blink,
                        pfn_obj.PteAddress,
                        pte_physical_address,
                        pfn_obj.u3.e2.ReferenceCount,
                        pfn_obj.u3.e1.m("PageColor") or
                        pfn_obj.u4.m("PageColor"),
                        containing_page,
                        pfn_obj.Type,
                        short_flags_string,
                        long_flags_string)


class PtoV(common.WinProcessFilter):
    """Converts a physical address to a virtual address."""

    __name = "ptov"

    PAGE_SIZE = 0x1000
    PAGE_BITS = 12

    @classmethod
    def args(cls, parser):
        super(PtoV, cls).args(parser)
        parser.add_argument("physical_address", type="IntParser",
                            help="The Virtual Address to examine.")

    def __init__(self, physical_address=None, **kwargs):
        """Converts a physical address to a virtual address."""
        super(PtoV, self).__init__(**kwargs)

        # Get a handle to the pfninfo plugin
        self.pfn_plugin = self.session.plugins.pfn(session=self.session)
        self.physical_address = physical_address

    def _ptov_x86(self, physical_address):
        """An implementation of ptov for x86."""
        result = physical_address & 0xFFF

        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 10) & 0x3FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 20) & 0xffc00000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def _ptov_x86_pae(self, physical_address):
        """An implementation of ptov for x86 pae."""
        result = physical_address & 0xFFF
        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 9) & 0x1FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 18) & 0x3fe00000

        # Get the PDPTE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject(
                "PDPTE invalid (Is this a one gig page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pdpte_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pdpte_address << 27) & 0x7FC0000000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PDPTE", pdpte_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def _ptov_x64(self, physical_address):
        """An implementation of ptov for x64."""
        result = physical_address & 0xFFF

        # Get the pte for this physical_address using the pfn database.
        pfn_obj = self.pfn_plugin.pfn_record(physical_address >> self.PAGE_BITS)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PTE invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pte_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pte_address << 9) & 0x1FF000

        # Get the PDE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PDE invalid (Is this a large page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pde_address = ((containing_page << self.PAGE_BITS) |
                       (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pde_address << 18) & 0x3fe00000

        # Get the PDPTE now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject(
                "PDPTE invalid (Is this a one gig page?)."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pdpte_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pdpte_address << 27) & 0x7FC0000000

        # Get the PML4E now:
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        if pfn_obj.Type != "ActiveAndValid":
            return obj.NoneObject("PML4E invalid."), []

        containing_page = int(pfn_obj.u4.PteFrame)
        pml4e_address = ((containing_page << self.PAGE_BITS) |
                         (int(pfn_obj.PteAddress) & 0xFFF))

        result |= (pml4e_address << 36) & 0xff8000000000

        # Now get the DTB.
        pfn_obj = self.pfn_plugin.pfn_record(containing_page)

        containing_page = int(pfn_obj.u4.PteFrame)
        dtb_address = containing_page << self.PAGE_BITS

        return result, (("DTB", dtb_address),
                        ("PML4E", pml4e_address),
                        ("PDPTE", pdpte_address),
                        ("PDE", pde_address),
                        ("PTE", pte_address))

    def ptov(self, physical_address):
        """Convert the physical address to a virtual address.

        Returns:
          a tuple (_EPROCESS of owning process, virtual address in process AS).
        """
        if self.profile.metadata("arch") == "I386":
            if self.profile.metadata("pae"):
                return self._ptov_x86_pae(physical_address)
            else:
                return self._ptov_x86(physical_address)
        elif self.profile.metadata("arch") == "AMD64":
            return self._ptov_x64(physical_address)

        return obj.NoneObject("Memory model not supported."), []

    def render(self, renderer):
        if self.physical_address is None:
            return

        result, structures = self.ptov(self.physical_address)
        if result:
            renderer.format("Physical Address {0:#x} => "
                            "Virtual Address {1:#x}\n",
                            self.physical_address, result)

            for type, phys_addr in structures:
                renderer.format("{0} @ {1:#x}\n", type, phys_addr)
        else:
            renderer.format("Error converting Physical Address {0:#x}: "
                            "{1}\n", self.physical_address, result)


class DTBScan2(common.WindowsCommandPlugin):
    """A Fast scanner for hidden DTBs.

    This scanner uses the fact that the virtual address of the DTB is always the
    same. We walk over all the physical pages, assume each page is a DTB and try
    to resolve the constant to a physical address.

    This plugin was written based on ideas and discussion with thomasdullien.
    """

    name = "dtbscan2"

    def TestVAddr(self, test_as, vaddr, symbol_checks):
        for vaddr, paddr in symbol_checks:
            if test_as.vtop(vaddr) != paddr:
                return False
        return True

    def render(self, renderer):
        dtb_map = {}
        pslist_plugin = self.session.plugins.pslist()
        for task in pslist_plugin.filter_processes():
            dtb = task.Pcb.DirectoryTableBase.v()
            dtb_map[dtb] = task

        symbols = ["nt", "nt!MmGetPhysicalMemoryRanges"]
        if self.session.profile.metadata("arch") == "AMD64":
            dtb_step = 0x1000
            # Add _KUSER_SHARED_DATA
            symbols.append(0xFFFFF78000000000)
        else:
            dtb_step = 0x20
            symbols.append(0xFFDF0000)

        symbol_checks = []
        for symbol in symbols:
            vaddr = self.session.address_resolver.get_address_by_name(symbol)
            paddr = self.session.kernel_address_space.vtop(vaddr)
            symbol_checks.append((vaddr, paddr))

        renderer.table_header([("DTB", "dtb", "[addrpad]"),
                               dict(name="Process", type="_EPROCESS"),
                              ])

        descriptor = self.profile.get_constant_object(
            "MmPhysicalMemoryBlock",
            target="Pointer",
            target_args=dict(
                target="_PHYSICAL_MEMORY_DESCRIPTOR",
                ))

        for memory_range in descriptor.Run:
            start = memory_range.BasePage * 0x1000
            length = memory_range.PageCount * 0x1000

            for page in range(start, start+length, dtb_step):
                self.session.report_progress("Checking %#x", page)
                test_as = self.session.kernel_address_space.__class__(
                    dtb=page, base=self.physical_address_space)

                if self.TestVAddr(test_as, vaddr, symbol_checks):
                    renderer.table_row(
                        page,
                        dtb_map.get(page, obj.NoneObject("Unknown"))
                    )


class DTBScan(common.WinProcessFilter):
    """Scans the physical memory for DTB values.

    This plugin can compare the DTBs found against the list of known processes
    to find hidden processes.
    """

    __name = "dtbscan"

    @classmethod
    def args(cls, parser):
        super(DTBScan, cls).args(parser)
        parser.add_argument("--limit", type="IntParser", default=0,
                            help="Stop scanning after this many mb.")

    def __init__(self, limit=None, **kwargs):
        super(DTBScan, self).__init__(**kwargs)
        self.limit = limit

    def render(self, renderer):
        ptov = self.session.plugins.ptov(session=self.session)
        pslist = self.session.plugins.pslist(session=self.session)
        pfn_plugin = self.session.plugins.pfn(session=self.session)

        # Known tasks:
        known_tasks = set()
        for task in pslist.list_eprocess():
            known_tasks.add(task.obj_offset)

        renderer.table_header([("DTB", "dtb", "[addrpad]"),
                               ("VAddr", "vaddr", "[addrpad]"),
                               dict(type="_EPROCESS"),
                               ("Known", "known", "")])

        seen_dtbs = set()

        # Now scan all the physical address space for DTBs.
        for _ in self.physical_address_space.get_available_addresses():
            start, _, length = _
            for page in range(start, start + length, 0x1000):
                self.session.report_progress("Scanning 0x%08X (%smb)" % (
                    page, page/1024/1024))

                # Quit early if requested to.
                if self.limit and page > self.limit:
                    return

                virtual_address, results = ptov.ptov(page)
                if virtual_address:
                    dtb = results[0][1]
                    if dtb not in seen_dtbs:
                        seen_dtbs.add(dtb)

                        # The _EPROCESS address is stored as the
                        # KernelStackOwner for the pfn of this dtb.
                        task = pfn_plugin.pfn_record(
                            dtb >> 12).u1.Flink.cast(
                                "Pointer", target="_EPROCESS").deref()

                        va, _ = ptov.ptov(dtb)
                        renderer.table_row(dtb, va, task,
                                           task.obj_offset in known_tasks)

class TestDTBScan(testlib.SimpleTestCase):
    PARAMETERS = dict(
        commandline="dtbscan --limit 10mb",
        )
